from typing import cast

from pytest import raises

from graphql import graphql_sync
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLFloat,
    GraphQLID,
    GraphQLInt,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
    assert_enum_type,
)
from graphql.utilities import (
    build_schema,
    build_client_schema,
    introspection_from_schema,
    print_schema,
)
from graphql.utilities.get_introspection_query import (
    IntrospectionEnumType,
    IntrospectionInputObjectType,
    IntrospectionInterfaceType,
    IntrospectionObjectType,
    IntrospectionType,
    IntrospectionUnionType,
)

from ..utils import dedent


def cycle_introspection(sdl_string: str):
    """Test that the client side introspection gives the same result.

    This function does a full cycle of going from a string with the contents of the SDL,
    build in-memory GraphQLSchema from it, produce a client-side representation of the
    schema by using "build_client_schema" and then return that schema printed as SDL.
    """
    server_schema = build_schema(sdl_string)
    initial_introspection = introspection_from_schema(server_schema)
    client_schema = build_client_schema(initial_introspection)
    # If the client then runs the introspection query against the client-side schema,
    # it should get a result identical to what was returned by the server
    second_introspection = introspection_from_schema(client_schema)

    # If the client then runs the introspection query against the client-side
    # schema, it should get a result identical to what was returned by the server.
    assert initial_introspection == second_introspection
    return print_schema(client_schema)


def describe_type_system_build_schema_from_introspection():
    def builds_a_simple_schema():
        sdl = dedent(
            '''
            """Simple schema"""
            schema {
              query: Simple
            }

            """This is a simple type"""
            type Simple {
              """This is a string field"""
              string: String
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_without_the_query_type():
        sdl = dedent(
            """
            type Query {
              foo: String
            }
            """
        )

        schema = build_schema(sdl)
        introspection = introspection_from_schema(schema)
        del introspection["__schema"]["queryType"]  # type: ignore

        client_schema = build_client_schema(introspection)
        assert client_schema.query_type is None
        assert print_schema(client_schema) == sdl

    def builds_a_simple_schema_with_all_operation_types():
        sdl = dedent(
            '''
            schema {
              query: QueryType
              mutation: MutationType
              subscription: SubscriptionType
            }

            """This is a simple mutation type"""
            type MutationType {
              """Set the string field"""
              string: String
            }

            """This is a simple query type"""
            type QueryType {
              """This is a string field"""
              string: String
            }

            """This is a simple subscription type"""
            type SubscriptionType {
              """This is a string field"""
              string: String
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def uses_built_in_scalars_when_possible():
        sdl = dedent(
            """
            scalar CustomScalar

            type Query {
              int: Int
              float: Float
              string: String
              boolean: Boolean
              id: ID
              custom: CustomScalar
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

        schema = build_schema(sdl)
        introspection = introspection_from_schema(schema)
        client_schema = build_client_schema(introspection)

        # Built-ins are used
        assert client_schema.get_type("Int") is GraphQLInt
        assert client_schema.get_type("Float") is GraphQLFloat
        assert client_schema.get_type("String") is GraphQLString
        assert client_schema.get_type("Boolean") is GraphQLBoolean
        assert client_schema.get_type("ID") is GraphQLID

        # Custom are built
        custom_scalar = schema.get_type("CustomScalar")
        assert client_schema.get_type("CustomScalar") is not custom_scalar

    def includes_standard_types_only_if_they_are_used():
        schema = build_schema(
            """
            type Query {
              foo: String
            }
            """
        )
        introspection = introspection_from_schema(schema)
        client_schema = build_client_schema(introspection)

        assert client_schema.get_type("Int") is None
        assert client_schema.get_type("Float") is None
        assert client_schema.get_type("ID") is None

    def builds_a_schema_with_a_recursive_type_reference():
        sdl = dedent(
            """
            schema {
              query: Recur
            }

            type Recur {
              recur: Recur
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_a_circular_type_reference():
        sdl = dedent(
            """
            type Dog {
              bestFriend: Human
            }

            type Human {
              bestFriend: Dog
            }

            type Query {
              dog: Dog
              human: Human
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_an_interface():
        sdl = dedent(
            '''
            type Dog implements Friendly {
              bestFriend: Friendly
            }

            interface Friendly {
              """The best friend of this friendly thing"""
              bestFriend: Friendly
            }

            type Human implements Friendly {
              bestFriend: Friendly
            }

            type Query {
              friendly: Friendly
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_an_interface_hierarchy():
        sdl = dedent(
            '''
            type Dog implements Friendly & Named {
              bestFriend: Friendly
              name: String
            }

            interface Friendly implements Named {
              """The best friend of this friendly thing"""
              bestFriend: Friendly
              name: String
            }

            type Human implements Friendly & Named {
              bestFriend: Friendly
              name: String
            }

            interface Named {
              name: String
            }

            type Query {
              friendly: Friendly
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_an_implicit_interface():
        sdl = dedent(
            '''
            type Dog implements Friendly {
              bestFriend: Friendly
            }

            interface Friendly {
              """The best friend of this friendly thing"""
              bestFriend: Friendly
            }

            type Query {
              dog: Dog
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_a_union():
        sdl = dedent(
            """
            type Dog {
              bestFriend: Friendly
            }

            union Friendly = Dog | Human

            type Human {
              bestFriend: Friendly
            }

            type Query {
              friendly: Friendly
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_complex_field_values():
        sdl = dedent(
            """
            type Query {
              string: String
              listOfString: [String]
              nonNullString: String!
              nonNullListOfString: [String]!
              nonNullListOfNonNullString: [String!]!
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_field_arguments():
        sdl = dedent(
            '''
            type Query {
              """A field with a single arg"""
              one(
                """This is an int arg"""
                intArg: Int
              ): String

              """A field with a two args"""
              two(
                """This is an list of int arg"""
                listArg: [Int]

                """This is a required arg"""
                requiredArg: Boolean!
              ): String
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_default_value_on_custom_scalar_field():
        sdl = dedent(
            """
            scalar CustomScalar

            type Query {
              testField(testArg: CustomScalar = "default"): String
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_an_enum():
        food_enum = GraphQLEnumType(
            "Food",
            {
                "VEGETABLES": GraphQLEnumValue(
                    1, description="Foods that are vegetables."
                ),
                "FRUITS": GraphQLEnumValue(2),
                "OILS": GraphQLEnumValue(3, deprecation_reason="Too fatty."),
            },
            description="Varieties of food stuffs",
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "EnumFields",
                {
                    "food": GraphQLField(
                        food_enum,
                        args={
                            "kind": GraphQLArgument(
                                food_enum, description="what kind of food?"
                            )
                        },
                        description="Repeats the arg you give it",
                    )
                },
            )
        )

        introspection = introspection_from_schema(schema)
        client_schema = build_client_schema(introspection)

        second_introspection = introspection_from_schema(client_schema)
        assert second_introspection == introspection

        # It's also an Enum type on the client.
        client_food_enum = assert_enum_type(client_schema.get_type("Food"))

        # Client types do not get server-only values, so the values mirror the names,
        # rather than using the integers defined in the "server" schema.
        values = {
            name: value.to_kwargs() for name, value in client_food_enum.values.items()
        }
        assert values == {
            "VEGETABLES": {
                "value": "VEGETABLES",
                "description": "Foods that are vegetables.",
                "deprecation_reason": None,
                "extensions": {},
                "ast_node": None,
            },
            "FRUITS": {
                "value": "FRUITS",
                "description": None,
                "deprecation_reason": None,
                "extensions": {},
                "ast_node": None,
            },
            "OILS": {
                "value": "OILS",
                "description": None,
                "deprecation_reason": "Too fatty.",
                "extensions": {},
                "ast_node": None,
            },
        }

    def builds_a_schema_with_an_input_object():
        sdl = dedent(
            '''
            """An input address"""
            input Address {
              """What street is this address?"""
              street: String!

              """The city the address is within?"""
              city: String!

              """The country (blank will assume USA)."""
              country: String = "USA"
            }

            type Query {
              """Get a geocode from an address"""
              geocode(
                """The address to lookup"""
                address: Address
              ): String
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_field_arguments_with_default_values():
        sdl = dedent(
            """
            input Geo {
              lat: Float
              lon: Float
            }

            type Query {
              defaultInt(intArg: Int = 30): String
              defaultList(listArg: [Int] = [1, 2, 3]): String
              defaultObject(objArg: Geo = {lat: 37.485, lon: -122.148}): String
              defaultNull(intArg: Int = null): String
              noDefault(intArg: Int): String
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_custom_directives():
        sdl = dedent(
            '''
            """This is a custom directive"""
            directive @customDirective repeatable on FIELD

            type Query {
              string: String
            }
            '''
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_without_directives():
        sdl = dedent(
            """
            type Query {
              foo: String
            }
            """
        )

        schema = build_schema(sdl)
        introspection = introspection_from_schema(schema)
        del introspection["__schema"]["directives"]  # type: ignore

        client_schema = build_client_schema(introspection)

        assert schema.directives
        assert client_schema.directives == ()
        assert print_schema(client_schema) == sdl

    def builds_a_schema_aware_of_deprecation():
        sdl = dedent(
            '''
            directive @someDirective(
              """This is a shiny new argument"""
              shinyArg: SomeInputObject

              """This was our design mistake :("""
              oldArg: String @deprecated(reason: "Use shinyArg")
            ) on QUERY

            enum Color {
              """So rosy"""
              RED

              """So grassy"""
              GREEN

              """So calming"""
              BLUE

              """So sickening"""
              MAUVE @deprecated(reason: "No longer in fashion")
            }

            input SomeInputObject {
              """Nothing special about it, just deprecated for some unknown reason"""
              oldField: String @deprecated(reason: "Don't use it, use newField instead!")

              """Same field but with a new name"""
              newField: String
            }

            type Query {
              """This is a shiny string field"""
              shinyString: String

              """This is a deprecated string field"""
              deprecatedString: String @deprecated(reason: "Use shinyString")

              """Color of a week"""
              color: Color

              """Some random field"""
              someField(
                """This is a shiny new argument"""
                shinyArg: SomeInputObject

                """This was our design mistake :("""
                oldArg: String @deprecated(reason: "Use shinyArg")
              ): String
            }
            '''  # noqa: E501
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_empty_deprecation_reasons():
        sdl = dedent(
            """
            directive @someDirective(someArg: SomeInputObject @deprecated(reason: "")) on QUERY

            type Query {
              someField(someArg: SomeInputObject @deprecated(reason: "")): SomeEnum @deprecated(reason: "")
            }

            input SomeInputObject {
              someInputField: String @deprecated(reason: "")
            }

            enum SomeEnum {
              SOME_VALUE @deprecated(reason: "")
            }
            """  # noqa: E501
        )

        assert cycle_introspection(sdl) == sdl

    def builds_a_schema_with_specified_by_url():
        sdl = dedent(
            """
            scalar Foo @specifiedBy(url: "https://example.com/foo_spec")

            type Query {
              foo: Foo
            }
            """
        )

        assert cycle_introspection(sdl) == sdl

    def can_use_client_schema_for_limited_execution():
        schema = build_schema(
            """
            scalar CustomScalar

            type Query {
              foo(custom1: CustomScalar, custom2: CustomScalar): String
            }
            """
        )

        introspection = introspection_from_schema(schema)
        client_schema = build_client_schema(introspection)

        class Data:
            foo = "bar"
            unused = "value"

        result = graphql_sync(
            client_schema,
            "query Limited($v: CustomScalar) { foo(custom1: 123, custom2: $v) }",
            root_value=Data(),
            variable_values={"v": "baz"},
        )

        assert result.data == {"foo": "bar"}

    def can_build_invalid_schema():
        schema = build_schema("type Query", assume_valid=True)

        introspection = introspection_from_schema(schema)
        client_schema = build_client_schema(introspection, assume_valid=True)

        assert client_schema.to_kwargs()["assume_valid"] is True

    def describe_throws_when_given_invalid_introspection():
        dummy_schema = build_schema(
            """
            type Query {
              foo(bar: String): String
            }

            interface SomeInterface {
              foo: String
            }

            union SomeUnion = Query

            enum SomeEnum { FOO }

            input SomeInputObject {
              foo: String
            }

            directive @SomeDirective on QUERY
            """
        )

        def throws_when_introspection_is_missing_schema_property():
            with raises(TypeError) as exc_info:
                # noinspection PyTypeChecker
                build_client_schema(None)  # type: ignore

            assert str(exc_info.value) == (
                "Invalid or incomplete introspection result. Ensure that you"
                " are passing the 'data' attribute of an introspection response"
                " and no 'errors' were returned alongside: None."
            )

            with raises(TypeError) as exc_info:
                # noinspection PyTypeChecker
                build_client_schema({})  # type: ignore

            assert str(exc_info.value) == (
                "Invalid or incomplete introspection result. Ensure that you"
                " are passing the 'data' attribute of an introspection response"
                " and no 'errors' were returned alongside: {}."
            )

        def throws_when_referenced_unknown_type():
            introspection = introspection_from_schema(dummy_schema)

            introspection["__schema"]["types"] = [
                type_
                for type_ in introspection["__schema"]["types"]
                if type_["name"] != "Query"
            ]

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value) == (
                "Invalid or incomplete schema, unknown type: Query."
                " Ensure that a full introspection query is used"
                " in order to build a client schema."
            )

        def throws_when_missing_definition_for_one_of_the_standard_scalars():
            schema = build_schema(
                """
                type Query {
                  foo: Float
                }
                """
            )
            introspection = introspection_from_schema(schema)
            introspection["__schema"]["types"] = [
                type_
                for type_ in introspection["__schema"]["types"]
                if type_["name"] != "Float"
            ]

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value).endswith(
                "Invalid or incomplete schema, unknown type: Float."
                " Ensure that a full introspection query is used"
                " in order to build a client schema."
            )

        def throws_when_type_reference_is_missing_name():
            introspection = introspection_from_schema(dummy_schema)
            query_type = cast(IntrospectionType, introspection["__schema"]["queryType"])
            assert query_type["name"] == "Query"
            del query_type["name"]  # type: ignore

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value) == "Unknown type reference: {}."

        def throws_when_missing_kind():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = next(
                type_
                for type_ in introspection["__schema"]["types"]
                if type_["name"] == "Query"
            )
            assert query_type_introspection["kind"] == "OBJECT"
            del query_type_introspection["kind"]

            with raises(
                TypeError,
                match=r"^Invalid or incomplete introspection result\."
                " Ensure that a full introspection query is used"
                r" in order to build a client schema: {'name': 'Query', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_interfaces():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Query"
                ),
            )

            assert query_type_introspection["interfaces"] == []
            del query_type_introspection["interfaces"]  # type: ignore

            with raises(
                TypeError,
                match="^Query interfaces cannot be resolved."
                " Introspection result missing interfaces:"
                r" {'kind': 'OBJECT', 'name': 'Query', .*}\.$",
            ):
                build_client_schema(introspection)

        def legacy_support_for_interfaces_with_null_as_interfaces_field():
            introspection = introspection_from_schema(dummy_schema)
            some_interface_introspection = cast(
                IntrospectionInterfaceType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "SomeInterface"
                ),
            )

            assert some_interface_introspection["interfaces"] == []
            some_interface_introspection["interfaces"] = None  # type: ignore

            client_schema = build_client_schema(introspection)
            assert print_schema(client_schema) == print_schema(dummy_schema)

        def throws_when_missing_fields():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Query"
                ),
            )

            assert query_type_introspection["fields"]
            del query_type_introspection["fields"]  # type: ignore

            with raises(
                TypeError,
                match="^Query fields cannot be resolved."
                " Introspection result missing fields:"
                r" {'kind': 'OBJECT', 'name': 'Query', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_field_args():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Query"
                ),
            )

            field = query_type_introspection["fields"][0]
            assert field["args"]
            del field["args"]  # type: ignore

            with raises(
                TypeError,
                match="^Query fields cannot be resolved."
                r" Introspection result missing field args: {'name': 'foo', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_output_type_is_used_as_an_arg_type():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Query"
                ),
            )

            arg = query_type_introspection["fields"][0]["args"][0]
            assert arg["type"]["name"] == "String"
            arg["type"]["name"] = "SomeUnion"

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value).startswith(
                "Query fields cannot be resolved."
                " Introspection must provide input type for arguments,"
                " but received: SomeUnion."
            )

        def throws_when_output_type_is_used_as_an_input_value_type():
            introspection = introspection_from_schema(dummy_schema)
            input_object_type_introspection = cast(
                IntrospectionInputObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "SomeInputObject"
                ),
            )

            input_field = input_object_type_introspection["inputFields"][0]
            assert input_field["type"]["name"] == "String"
            input_field["type"]["name"] = "SomeUnion"

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value).startswith(
                "SomeInputObject fields cannot be resolved."
                " Introspection must provide input type for input fields,"
                " but received: SomeUnion."
            )

        def throws_when_input_type_is_used_as_a_field_type():
            introspection = introspection_from_schema(dummy_schema)
            query_type_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Query"
                ),
            )

            field = query_type_introspection["fields"][0]
            assert field["type"]["name"] == "String"
            field["type"]["name"] = "SomeInputObject"

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value).startswith(
                "Query fields cannot be resolved."
                " Introspection must provide output type for fields,"
                " but received: SomeInputObject."
            )

        def throws_when_missing_possible_types():
            introspection = introspection_from_schema(dummy_schema)
            some_union_introspection = cast(
                IntrospectionUnionType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "SomeUnion"
                ),
            )

            assert some_union_introspection["possibleTypes"]
            del some_union_introspection["possibleTypes"]  # type: ignore

            with raises(
                TypeError,
                match="^Introspection result missing possibleTypes:"
                r" {'kind': 'UNION', 'name': 'SomeUnion', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_enum_values():
            introspection = introspection_from_schema(dummy_schema)
            some_enum_introspection = cast(
                IntrospectionEnumType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "SomeEnum"
                ),
            )

            assert some_enum_introspection["enumValues"]
            del some_enum_introspection["enumValues"]  # type: ignore

            with raises(
                TypeError,
                match="^Introspection result missing enumValues:"
                r" {'kind': 'ENUM', 'name': 'SomeEnum', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_input_fields():
            introspection = introspection_from_schema(dummy_schema)
            some_input_object_introspection = cast(
                IntrospectionInputObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "SomeInputObject"
                ),
            )

            assert some_input_object_introspection["inputFields"]
            del some_input_object_introspection["inputFields"]  # type: ignore

            with raises(
                TypeError,
                match="^Introspection result missing inputFields:"
                r" {'kind': 'INPUT_OBJECT', 'name': 'SomeInputObject', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_directive_locations():
            introspection = introspection_from_schema(dummy_schema)
            some_directive_introspection = introspection["__schema"]["directives"][0]

            assert some_directive_introspection["name"] == "SomeDirective"
            assert some_directive_introspection["locations"] == ["QUERY"]
            del some_directive_introspection["locations"]  # type: ignore

            with raises(
                TypeError,
                match="^Introspection result missing directive locations:"
                r" {'name': 'SomeDirective', .*}\.$",
            ):
                build_client_schema(introspection)

        def throws_when_missing_directive_args():
            introspection = introspection_from_schema(dummy_schema)
            some_directive_introspection = introspection["__schema"]["directives"][0]

            assert some_directive_introspection["name"] == "SomeDirective"
            assert some_directive_introspection["args"] == []
            del some_directive_introspection["args"]  # type: ignore

            with raises(
                TypeError,
                match="^Introspection result missing directive args:"
                r" {'name': 'SomeDirective', .*}\.$",
            ):
                build_client_schema(introspection)

    def describe_very_deep_decorators_are_not_supported():
        def fails_on_very_deep_lists_more_than_7_levels():
            schema = build_schema(
                """
                type Query {
                  foo: [[[[[[[[String]]]]]]]]
                }
                """
            )

            introspection = introspection_from_schema(schema)

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value) == (
                "Query fields cannot be resolved."
                " Decorated type deeper than introspection query."
            )

        def fails_on_a_very_deep_non_null_more_than_7_levels():
            schema = build_schema(
                """
                type Query {
                  foo: [[[[String!]!]!]!]
                }
                """
            )

            introspection = introspection_from_schema(schema)

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)

            assert str(exc_info.value) == (
                "Query fields cannot be resolved."
                " Decorated type deeper than introspection query."
            )

        def succeeds_on_deep_types_less_or_equal_7_levels():
            # e.g., fully non-null 3D matrix
            sdl = dedent(
                """
                type Query {
                  foo: [[[String!]!]!]!
                }
                """
            )

            assert cycle_introspection(sdl) == sdl

    def describe_prevents_infinite_recursion_on_invalid_introspection():
        def recursive_interfaces():
            sdl = """
                type Query {
                  foo: Foo
                }

                type Foo {
                  foo: String
                }
                """
            schema = build_schema(sdl, assume_valid=True)
            introspection = introspection_from_schema(schema)
            foo_introspection = cast(
                IntrospectionObjectType,
                next(
                    type_
                    for type_ in introspection["__schema"]["types"]
                    if type_["name"] == "Foo"
                ),
            )

            assert foo_introspection["interfaces"] == []
            # we need to patch here since invalid interfaces cannot be built with Python
            foo_introspection["interfaces"] = [
                {"kind": "OBJECT", "name": "Foo", "ofType": None}
            ]

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)
            assert str(exc_info.value) == (
                "Foo interfaces cannot be resolved."
                " Expected Foo to be a GraphQL Interface type."
            )

        def recursive_union():
            sdl = """
                type Query {
                  foo: Foo
                }

                union Foo
                """
            schema = build_schema(sdl, assume_valid=True)
            introspection = introspection_from_schema(schema)
            foo_introspection = next(
                type_
                for type_ in introspection["__schema"]["types"]
                if type_["name"] == "Foo"
            )

            assert foo_introspection["kind"] == "UNION"
            assert foo_introspection["possibleTypes"] == []
            # we need to patch here since invalid unions cannot be built with Python
            foo_introspection["possibleTypes"] = [
                {"kind": "UNION", "name": "Foo", "ofType": None}
            ]

            with raises(TypeError) as exc_info:
                build_client_schema(introspection)
            assert str(exc_info.value) == (
                "Foo types cannot be resolved."
                " Expected Foo to be a GraphQL Object type."
            )



from typing import Any, List, NamedTuple, Optional, Union

__all__ = ["Path"]


class Path(NamedTuple):
    """A generic path of string or integer indices"""

    prev: Any  # Optional['Path'] (python/mypy/issues/731)
    """path with the previous indices"""
    key: Union[str, int]
    """current index in the path (string or integer)"""
    typename: Optional[str]
    """name of the parent type to avoid path ambiguity"""

    def add_key(self, key: Union[str, int], typename: Optional[str] = None) -> "Path":
        """Return a new Path containing the given key."""
        return Path(self, key, typename)

    def as_list(self) -> List[Union[str, int]]:
        """Return a list of the path keys."""
        flattened: List[Union[str, int]] = []
        append = flattened.append
        curr: Path = self
        while curr:
            append(curr.key)
            curr = curr.prev
        return flattened[::-1]



from functools import partial

from graphql.validation import UniqueFragmentNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, UniqueFragmentNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_fragment_names():
    def no_fragments():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def one_fragment():
        assert_valid(
            """
            {
              ...fragA
            }
            fragment fragA on Type {
              field
            }
            """
        )

    def many_fragments():
        assert_valid(
            """
            {
              ...fragA
              ...fragB
              ...fragC
            }
            fragment fragA on Type {
              fieldA
            }
            fragment fragB on Type {
              fieldB
            }
            fragment fragC on Type {
              fieldC
            }
            """
        )

    def inline_fragments_are_always_unique():
        assert_valid(
            """
            {
              ...on Type {
                fieldA
              }
              ...on Type {
                fieldB
              }
            }
            """
        )

    def fragment_and_operation_named_the_same():
        assert_valid(
            """
            query Foo {
              ...Foo
            }
            fragment Foo on Type {
              field
            }
            """
        )

    def fragments_named_the_same():
        assert_errors(
            """
            {
              ...fragA
            }
            fragment fragA on Type {
              fieldA
            }
            fragment fragA on Type {
              fieldB
            }
            """,
            [
                {
                    "message": "There can be only one fragment named 'fragA'.",
                    "locations": [(5, 22), (8, 22)],
                },
            ],
        )

    def fragments_named_the_same_without_being_referenced():
        assert_errors(
            """
            fragment fragA on Type {
              fieldA
            }
            fragment fragA on Type {
              fieldB
            }
            """,
            [
                {
                    "message": "There can be only one fragment named 'fragA'.",
                    "locations": [(2, 22), (5, 22)],
                },
            ],
        )



from graphql.pyutils import did_you_mean


def describe_did_you_mean():
    def does_accept_an_empty_list():
        assert did_you_mean([]) == ""

    def handles_single_suggestion():
        assert did_you_mean(["A"]) == " Did you mean 'A'?"

    def handles_two_suggestions():
        assert did_you_mean(["A", "B"]) == " Did you mean 'A' or 'B'?"

    def handles_multiple_suggestions():
        assert did_you_mean(["A", "B", "C"]) == " Did you mean 'A', 'B', or 'C'?"

    def limits_to_five_suggestions():
        assert (
            did_you_mean(["A", "B", "C", "D", "E", "F"])
            == " Did you mean 'A', 'B', 'C', 'D', or 'E'?"
        )

    def adds_sub_message():
        assert did_you_mean(["A"], "the letter") == " Did you mean the letter 'A'?"



import asyncio

from typing import Any, Dict, List, Callable

from pytest import mark, raises

from graphql.language import parse
from graphql.pyutils import SimplePubSub
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLField,
    GraphQLInt,
    GraphQLList,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)
from graphql.execution import create_source_event_stream, subscribe, MapAsyncIterator

try:
    anext
except NameError:  # pragma: no cover (Python < 3.10)
    # noinspection PyShadowingBuiltins
    async def anext(iterator):
        """Return the next item from an async iterator."""
        return await iterator.__anext__()


Email = Dict  # should become a TypedDict once we require Python 3.8

EmailType = GraphQLObjectType(
    "Email",
    {
        "from": GraphQLField(GraphQLString),
        "subject": GraphQLField(GraphQLString),
        "message": GraphQLField(GraphQLString),
        "unread": GraphQLField(GraphQLBoolean),
    },
)

InboxType = GraphQLObjectType(
    "Inbox",
    {
        "total": GraphQLField(
            GraphQLInt, resolve=lambda inbox, _info: len(inbox["emails"])
        ),
        "unread": GraphQLField(
            GraphQLInt,
            resolve=lambda inbox, _info: sum(
                1 for email in inbox["emails"] if email["unread"]
            ),
        ),
        "emails": GraphQLField(GraphQLList(EmailType)),
    },
)

QueryType = GraphQLObjectType("Query", {"inbox": GraphQLField(InboxType)})

EmailEventType = GraphQLObjectType(
    "EmailEvent", {"email": GraphQLField(EmailType), "inbox": GraphQLField(InboxType)}
)


email_schema = GraphQLSchema(
    query=QueryType,
    subscription=GraphQLObjectType(
        "Subscription",
        {
            "importantEmail": GraphQLField(
                EmailEventType,
                args={"priority": GraphQLArgument(GraphQLInt)},
            )
        },
    ),
)


def create_subscription(pubsub: SimplePubSub):
    document = parse(
        """
        subscription ($priority: Int = 0) {
          importantEmail(priority: $priority) {
            email {
              from
              subject
            }
            inbox {
              unread
              total
            }
          }
        }
        """
    )

    emails: List[Email] = [
        {
            "from": "joe@graphql.org",
            "subject": "Hello",
            "message": "Hello World",
            "unread": False,
        }
    ]

    def transform(new_email):
        emails.append(new_email)

        return {"importantEmail": {"email": new_email, "inbox": data["inbox"]}}

    data: Dict[str, Any] = {
        "inbox": {"emails": emails},
        "importantEmail": pubsub.get_subscriber(transform),
    }

    return subscribe(email_schema, document, data)


DummyQueryType = GraphQLObjectType("Query", {"dummy": GraphQLField(GraphQLString)})


# Check all error cases when initializing the subscription.
def describe_subscription_initialization_phase():
    @mark.asyncio
    async def accepts_positional_arguments():
        document = parse(
            """
            subscription {
              importantEmail
            }
            """
        )

        async def empty_async_iterator(_info):
            for value in ():  # type: ignore
                yield value  # pragma: no cover

        ai = await subscribe(
            email_schema, document, {"importantEmail": empty_async_iterator}
        )

        with raises(StopAsyncIteration):
            await anext(ai)
        await ai.aclose()  # type: ignore

    @mark.asyncio
    async def accepts_multiple_subscription_fields_defined_in_schema():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "foo": GraphQLField(GraphQLString),
                    "bar": GraphQLField(GraphQLString),
                },
            ),
        )

        async def foo_generator(_info):
            yield {"foo": "FooValue"}

        subscription = await subscribe(
            schema, parse("subscription { foo }"), {"foo": foo_generator}
        )
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == ({"foo": "FooValue"}, None)

        await subscription.aclose()

    @mark.asyncio
    async def accepts_type_definition_with_sync_subscribe_function():
        async def foo_generator(_obj, _info):
            yield {"foo": "FooValue"}

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {"foo": GraphQLField(GraphQLString, subscribe=foo_generator)},
            ),
        )

        subscription = await subscribe(schema, parse("subscription { foo }"))
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == ({"foo": "FooValue"}, None)

        await subscription.aclose()

    @mark.asyncio
    async def accepts_type_definition_with_async_subscribe_function():
        async def foo_generator(_obj, _info):
            await asyncio.sleep(0)
            yield {"foo": "FooValue"}

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {"foo": GraphQLField(GraphQLString, subscribe=foo_generator)},
            ),
        )

        subscription = await subscribe(schema, parse("subscription { foo }"))
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == ({"foo": "FooValue"}, None)

        await subscription.aclose()

    @mark.asyncio
    async def uses_a_custom_default_subscribe_field_resolver():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription", {"foo": GraphQLField(GraphQLString)}
            ),
        )

        class Root:
            @staticmethod
            async def custom_foo():
                yield {"foo": "FooValue"}

        subscription = await subscribe(
            schema,
            document=parse("subscription { foo }"),
            root_value=Root(),
            subscribe_field_resolver=lambda root, _info: root.custom_foo(),
        )
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == (
            {"foo": "FooValue"},
            None,
        )

        await subscription.aclose()

    @mark.asyncio
    async def should_only_resolve_the_first_field_of_invalid_multi_field():
        did_resolve = {"foo": False, "bar": False}

        async def subscribe_foo(_obj, _info):
            did_resolve["foo"] = True
            yield {"foo": "FooValue"}

        async def subscribe_bar(_obj, _info):  # pragma: no cover
            did_resolve["bar"] = True
            yield {"bar": "BarValue"}

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "foo": GraphQLField(GraphQLString, subscribe=subscribe_foo),
                    "bar": GraphQLField(GraphQLString, subscribe=subscribe_bar),
                },
            ),
        )

        subscription = await subscribe(schema, parse("subscription { foo bar }"))
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == (
            {"foo": "FooValue", "bar": None},
            None,
        )

        assert did_resolve == {"foo": True, "bar": False}

        await subscription.aclose()

    @mark.asyncio
    async def throws_an_error_if_some_of_required_arguments_are_missing():
        document = parse("subscription { foo }")

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription", {"foo": GraphQLField(GraphQLString)}
            ),
        )

        with raises(TypeError, match="^Expected None to be a GraphQL schema\\.$"):
            await subscribe(None, document)  # type: ignore

        with raises(TypeError, match="missing .* positional argument: 'schema'"):
            await subscribe(document=document)  # type: ignore

        with raises(TypeError, match="^Must provide document\\.$"):
            await subscribe(schema, None)  # type: ignore

        with raises(TypeError, match="missing .* positional argument: 'document'"):
            await subscribe(schema=schema)  # type: ignore

    @mark.asyncio
    async def resolves_to_an_error_if_schema_does_not_support_subscriptions():
        schema = GraphQLSchema(query=DummyQueryType)
        document = parse("subscription { unknownField }")

        result = await subscribe(schema, document)

        assert result == (
            None,
            [
                {
                    "message": "Schema is not configured to execute"
                    " subscription operation.",
                    "locations": [(1, 1)],
                }
            ],
        )

    @mark.asyncio
    async def resolves_to_an_error_for_unknown_subscription_field():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription", {"foo": GraphQLField(GraphQLString)}
            ),
        )
        document = parse("subscription { unknownField }")

        result = await subscribe(schema, document)
        assert result == (
            None,
            [
                {
                    "message": "The subscription field 'unknownField' is not defined.",
                    "locations": [(1, 16)],
                }
            ],
        )

    @mark.asyncio
    async def should_pass_through_unexpected_errors_thrown_in_subscribe():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription", {"foo": GraphQLField(GraphQLString)}
            ),
        )
        with raises(TypeError, match="^Must provide document\\.$"):
            await subscribe(schema=schema, document={})  # type: ignore

    @mark.asyncio
    @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
    async def throws_an_error_if_subscribe_does_not_return_an_iterator():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "foo": GraphQLField(
                        GraphQLString, subscribe=lambda _obj, _info: "test"
                    )
                },
            ),
        )

        document = parse("subscription { foo }")

        with raises(TypeError) as exc_info:
            await subscribe(schema, document)

        assert str(exc_info.value) == (
            "Subscription field must return AsyncIterable. Received: 'test'."
        )

    @mark.asyncio
    async def resolves_to_an_error_for_subscription_resolver_errors():
        async def subscribe_with_fn(subscribe_fn: Callable):
            schema = GraphQLSchema(
                query=DummyQueryType,
                subscription=GraphQLObjectType(
                    "Subscription",
                    {"foo": GraphQLField(GraphQLString, subscribe=subscribe_fn)},
                ),
            )
            document = parse("subscription { foo }")
            result = await subscribe(schema, document)

            assert await create_source_event_stream(schema, document) == result
            return result

        expected_result = (
            None,
            [
                {
                    "message": "test error",
                    "locations": [(1, 16)],
                    "path": ["foo"],
                }
            ],
        )

        # Returning an error
        def return_error(_obj, _info):
            return TypeError("test error")

        assert await subscribe_with_fn(return_error) == expected_result

        # Throwing an error
        def throw_error(*_args):
            raise TypeError("test error")

        assert await subscribe_with_fn(throw_error) == expected_result

        # Resolving to an error
        async def resolve_error(*_args):
            return TypeError("test error")

        assert await subscribe_with_fn(resolve_error) == expected_result

        # Rejecting with an error
        async def reject_error(*_args):
            return TypeError("test error")

        assert await subscribe_with_fn(reject_error) == expected_result

    @mark.asyncio
    async def resolves_to_an_error_if_variables_were_wrong_type():
        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "foo": GraphQLField(
                        GraphQLString, {"arg": GraphQLArgument(GraphQLInt)}
                    )
                },
            ),
        )

        variable_values = {"arg": "meow"}
        document = parse(
            """
            subscription ($arg: Int) {
              foo(arg: $arg)
            }
            """
        )

        # If we receive variables that cannot be coerced correctly, subscribe() will
        # resolve to an ExecutionResult that contains an informative error description.
        result = await subscribe(schema, document, variable_values=variable_values)

        assert result == (
            None,
            [
                {
                    "message": "Variable '$arg' got invalid value 'meow';"
                    " Int cannot represent non-integer value: 'meow'",
                    "locations": [(2, 27)],
                }
            ],
        )

        assert result.errors[0].original_error is None  # type: ignore


# Once a subscription returns a valid AsyncIterator, it can still yield errors.
def describe_subscription_publish_phase():
    @mark.asyncio
    async def produces_a_payload_for_multiple_subscribe_in_same_subscription():
        pubsub = SimplePubSub()

        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        second_subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        payload1 = anext(subscription)
        payload2 = anext(second_subscription)

        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        expected_payload = {
            "importantEmail": {
                "email": {"from": "yuzhi@graphql.org", "subject": "Alright"},
                "inbox": {"unread": 1, "total": 2},
            }
        }

        assert await payload1 == (expected_payload, None)
        assert await payload2 == (expected_payload, None)

    @mark.asyncio
    async def produces_a_payload_per_subscription_event():
        pubsub = SimplePubSub()
        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        # Wait for the next subscription payload.
        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        # The previously waited on payload now has a value.
        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Alright"},
                    "inbox": {"unread": 1, "total": 2},
                }
            },
            None,
        )

        # Another new email arrives, before anext(subscription) is called.
        assert (
            pubsub.emit(
                {
                    "from": "hyo@graphql.org",
                    "subject": "Tools",
                    "message": "I <3 making things",
                    "unread": True,
                }
            )
            is True
        )

        # The next waited on payload will have a value.
        assert await anext(subscription) == (
            {
                "importantEmail": {
                    "email": {"from": "hyo@graphql.org", "subject": "Tools"},
                    "inbox": {"unread": 2, "total": 3},
                }
            },
            None,
        )

        # The client decides to disconnect.
        # noinspection PyUnresolvedReferences
        await subscription.aclose()

        # Which may result in disconnecting upstream services as well.
        assert (
            pubsub.emit(
                {
                    "from": "adam@graphql.org",
                    "subject": "Important",
                    "message": "Read me please",
                    "unread": True,
                }
            )
            is False
        )  # No more listeners.

        # Awaiting subscription after closing it results in completed results.
        with raises(StopAsyncIteration):
            assert await anext(subscription)

    @mark.asyncio
    async def produces_a_payload_when_there_are_multiple_events():
        pubsub = SimplePubSub()
        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Alright"},
                    "inbox": {"unread": 1, "total": 2},
                }
            },
            None,
        )

        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright 2",
                    "message": "Tests are good 2",
                    "unread": True,
                }
            )
            is True
        )

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Alright 2"},
                    "inbox": {"unread": 2, "total": 3},
                }
            },
            None,
        )

    @mark.asyncio
    async def should_not_trigger_when_subscription_is_already_done():
        pubsub = SimplePubSub()
        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Alright"},
                    "inbox": {"unread": 1, "total": 2},
                }
            },
            None,
        )

        payload = anext(subscription)
        await subscription.aclose()

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright 2",
                    "message": "Tests are good 2",
                    "unread": True,
                }
            )
            is False
        )

        with raises(StopAsyncIteration):
            await payload

    @mark.asyncio
    async def should_not_trigger_when_subscription_is_thrown():
        pubsub = SimplePubSub()
        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Alright",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Alright"},
                    "inbox": {"unread": 1, "total": 2},
                }
            },
            None,
        )

        payload = anext(subscription)

        # Throw error
        with raises(RuntimeError) as exc_info:
            await subscription.athrow(RuntimeError("ouch"))
        assert str(exc_info.value) == "ouch"

        with raises(StopAsyncIteration):
            await payload

    @mark.asyncio
    async def event_order_is_correct_for_multiple_publishes():
        pubsub = SimplePubSub()
        subscription = await create_subscription(pubsub)
        assert isinstance(subscription, MapAsyncIterator)

        payload = anext(subscription)

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Message",
                    "message": "Tests are good",
                    "unread": True,
                }
            )
            is True
        )

        # A new email arrives!
        assert (
            pubsub.emit(
                {
                    "from": "yuzhi@graphql.org",
                    "subject": "Message 2",
                    "message": "Tests are good 2",
                    "unread": True,
                }
            )
            is True
        )

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Message"},
                    "inbox": {"unread": 2, "total": 3},
                }
            },
            None,
        )

        payload = anext(subscription)

        assert await payload == (
            {
                "importantEmail": {
                    "email": {"from": "yuzhi@graphql.org", "subject": "Message 2"},
                    "inbox": {"unread": 2, "total": 3},
                }
            },
            None,
        )

    @mark.asyncio
    async def should_handle_error_during_execution_of_source_event():
        async def generate_messages(_obj, _info):
            yield "Hello"
            yield "Goodbye"
            yield "Bonjour"

        def resolve_message(message, _info):
            if message == "Goodbye":
                raise RuntimeError("Never leave.")
            return message

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "newMessage": GraphQLField(
                        GraphQLString,
                        subscribe=generate_messages,
                        resolve=resolve_message,
                    )
                },
            ),
        )

        document = parse("subscription { newMessage }")
        subscription = await subscribe(schema, document)
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == ({"newMessage": "Hello"}, None)

        # An error in execution is presented as such.
        assert await anext(subscription) == (
            {"newMessage": None},
            [
                {
                    "message": "Never leave.",
                    "locations": [(1, 16)],
                    "path": ["newMessage"],
                }
            ],
        )

        # However that does not close the response event stream.
        # Subsequent events are still executed.
        assert await anext(subscription) == ({"newMessage": "Bonjour"}, None)

    @mark.asyncio
    async def should_pass_through_error_thrown_in_source_event_stream():
        async def generate_messages(_obj, _info):
            yield "Hello"
            raise RuntimeError("test error")

        def resolve_message(message, _info):
            return message

        schema = GraphQLSchema(
            query=DummyQueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "newMessage": GraphQLField(
                        GraphQLString,
                        resolve=resolve_message,
                        subscribe=generate_messages,
                    )
                },
            ),
        )

        document = parse("subscription { newMessage }")
        subscription = await subscribe(schema, document)
        assert isinstance(subscription, MapAsyncIterator)

        assert await (anext(subscription)) == ({"newMessage": "Hello"}, None)

        with raises(RuntimeError) as exc_info:
            await anext(subscription)

        assert str(exc_info.value) == "test error"

        with raises(StopAsyncIteration):
            await anext(subscription)

    @mark.asyncio
    async def should_work_with_async_resolve_function():
        async def generate_messages(_obj, _info):
            yield "Hello"

        def resolve_message(message, _info):
            return message

        schema = GraphQLSchema(
            query=QueryType,
            subscription=GraphQLObjectType(
                "Subscription",
                {
                    "newMessage": GraphQLField(
                        GraphQLString,
                        resolve=resolve_message,
                        subscribe=generate_messages,
                    )
                },
            ),
        )

        document = parse("subscription { newMessage }")
        subscription = await subscribe(schema, document)
        assert isinstance(subscription, MapAsyncIterator)

        assert await anext(subscription) == ({"newMessage": "Hello"}, None)



from typing import Any

from graphql import graphql_sync

from .star_wars_schema import star_wars_schema


def query_star_wars(source: str) -> Any:
    result = graphql_sync(star_wars_schema, source)
    assert result.errors is None
    return result.data


def describe_star_wars_introspection_tests():
    def describe_basic_introspection():
        def allows_querying_the_schema_for_types():
            data = query_star_wars(
                """
                {
                  __schema {
                    types {
                      name
                    }
                  }
                }
                """
            )
            # Include all types used by StarWars schema, introspection types and
            # standard directives. For example, `Boolean` is used in `@skip`,
            # `@include` and also inside introspection types.
            assert data == {
                "__schema": {
                    "types": [
                        {"name": "Human"},
                        {"name": "Character"},
                        {"name": "String"},
                        {"name": "Episode"},
                        {"name": "Droid"},
                        {"name": "Query"},
                        {"name": "Boolean"},
                        {"name": "__Schema"},
                        {"name": "__Type"},
                        {"name": "__TypeKind"},
                        {"name": "__Field"},
                        {"name": "__InputValue"},
                        {"name": "__EnumValue"},
                        {"name": "__Directive"},
                        {"name": "__DirectiveLocation"},
                    ]
                }
            }

        def allows_querying_the_schema_for_query_type():
            data = query_star_wars(
                """
                {
                  __schema {
                    queryType {
                      name
                    }
                  }
                }
                """
            )

            assert data == {"__schema": {"queryType": {"name": "Query"}}}

        def allows_querying_the_schema_for_a_specific_type():
            data = query_star_wars(
                """
                {
                  __type(name: "Droid") {
                    name
                  }
                }
                """
            )
            assert data == {"__type": {"name": "Droid"}}

        def allows_querying_the_schema_for_an_object_kind():
            data = query_star_wars(
                """
                {
                  __type(name: "Droid") {
                    name
                    kind
                  }
                }
                """
            )
            assert data == {"__type": {"name": "Droid", "kind": "OBJECT"}}

        def allows_querying_the_schema_for_an_interface_kind():
            data = query_star_wars(
                """
                {
                  __type(name: "Character") {
                    name
                    kind
                  }
                }
                """
            )
            assert data == {"__type": {"name": "Character", "kind": "INTERFACE"}}

        def allows_querying_the_schema_for_object_fields():
            data = query_star_wars(
                """
                {
                  __type(name: "Droid") {
                    name
                    fields {
                      name
                      type {
                        name
                        kind
                      }
                    }
                  }
                }
                """
            )
            assert data == {
                "__type": {
                    "name": "Droid",
                    "fields": [
                        {"name": "id", "type": {"name": None, "kind": "NON_NULL"}},
                        {"name": "name", "type": {"name": "String", "kind": "SCALAR"}},
                        {"name": "friends", "type": {"name": None, "kind": "LIST"}},
                        {"name": "appearsIn", "type": {"name": None, "kind": "LIST"}},
                        {
                            "name": "secretBackstory",
                            "type": {"name": "String", "kind": "SCALAR"},
                        },
                        {
                            "name": "primaryFunction",
                            "type": {"name": "String", "kind": "SCALAR"},
                        },
                    ],
                }
            }

        def allows_querying_the_schema_for_nested_object_fields():
            data = query_star_wars(
                """
                {
                  __type(name: "Droid") {
                    name
                    fields {
                      name
                      type {
                        name
                        kind
                        ofType {
                          name
                          kind
                        }
                      }
                    }
                  }
                }
                """
            )
            assert data == {
                "__type": {
                    "name": "Droid",
                    "fields": [
                        {
                            "name": "id",
                            "type": {
                                "name": None,
                                "kind": "NON_NULL",
                                "ofType": {"name": "String", "kind": "SCALAR"},
                            },
                        },
                        {
                            "name": "name",
                            "type": {
                                "name": "String",
                                "kind": "SCALAR",
                                "ofType": None,
                            },
                        },
                        {
                            "name": "friends",
                            "type": {
                                "name": None,
                                "kind": "LIST",
                                "ofType": {"name": "Character", "kind": "INTERFACE"},
                            },
                        },
                        {
                            "name": "appearsIn",
                            "type": {
                                "name": None,
                                "kind": "LIST",
                                "ofType": {"name": "Episode", "kind": "ENUM"},
                            },
                        },
                        {
                            "name": "secretBackstory",
                            "type": {
                                "name": "String",
                                "kind": "SCALAR",
                                "ofType": None,
                            },
                        },
                        {
                            "name": "primaryFunction",
                            "type": {
                                "name": "String",
                                "kind": "SCALAR",
                                "ofType": None,
                            },
                        },
                    ],
                }
            }

        def allows_querying_the_schema_for_field_args():
            data = query_star_wars(
                """
                {
                  __schema {
                    queryType {
                      fields {
                        name
                        args {
                          name
                          description
                          type {
                            name
                            kind
                            ofType {
                              name
                              kind
                            }
                          }
                          defaultValue
                        }
                      }
                    }
                  }
                }
                """
            )

            assert data == {
                "__schema": {
                    "queryType": {
                        "fields": [
                            {
                                "name": "hero",
                                "args": [
                                    {
                                        "defaultValue": None,
                                        "description": "If omitted, returns the hero of"
                                        " the whole saga. If provided, returns the hero"
                                        " of that particular episode.",
                                        "name": "episode",
                                        "type": {
                                            "kind": "ENUM",
                                            "name": "Episode",
                                            "ofType": None,
                                        },
                                    }
                                ],
                            },
                            {
                                "name": "human",
                                "args": [
                                    {
                                        "name": "id",
                                        "description": "id of the human",
                                        "type": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "SCALAR",
                                                "name": "String",
                                            },
                                        },
                                        "defaultValue": None,
                                    }
                                ],
                            },
                            {
                                "name": "droid",
                                "args": [
                                    {
                                        "name": "id",
                                        "description": "id of the droid",
                                        "type": {
                                            "kind": "NON_NULL",
                                            "name": None,
                                            "ofType": {
                                                "kind": "SCALAR",
                                                "name": "String",
                                            },
                                        },
                                        "defaultValue": None,
                                    }
                                ],
                            },
                        ]
                    }
                }
            }

        def allows_querying_the_schema_for_documentation():
            data = query_star_wars(
                """
                {
                  __type(name: "Droid") {
                    name
                    description
                  }
                }
                """
            )

            assert data == {
                "__type": {
                    "name": "Droid",
                    "description": "A mechanical creature in the Star Wars universe.",
                }
            }



from gc import collect
from inspect import isawaitable

from pytest import mark, raises

from graphql import graphql_sync
from graphql.execution import execute, execute_sync
from graphql.language import parse
from graphql.type import GraphQLField, GraphQLObjectType, GraphQLSchema, GraphQLString
from graphql.validation import validate


def describe_execute_synchronously_when_possible():
    def _resolve_sync(root_value, _info):
        return root_value

    async def _resolve_async(root_value, _info):
        return root_value

    schema = GraphQLSchema(
        GraphQLObjectType(
            "Query",
            {
                "syncField": GraphQLField(GraphQLString, resolve=_resolve_sync),
                "asyncField": GraphQLField(GraphQLString, resolve=_resolve_async),
            },
        ),
        GraphQLObjectType(
            "Mutation",
            {"syncMutationField": GraphQLField(GraphQLString, resolve=_resolve_sync)},
        ),
    )

    def does_not_return_an_awaitable_for_initial_errors():
        doc = "fragment Example on Query { syncField }"
        assert execute(schema, parse(doc), "rootValue") == (
            None,
            [{"message": "Must provide an operation."}],
        )

    def does_not_return_an_awaitable_if_fields_are_all_synchronous():
        doc = "query Example { syncField }"
        assert execute(schema, parse(doc), "rootValue") == (
            {"syncField": "rootValue"},
            None,
        )

    def does_not_return_an_awaitable_if_mutation_fields_are_all_synchronous():
        doc = "mutation Example { syncMutationField }"
        assert execute(schema, parse(doc), "rootValue") == (
            {"syncMutationField": "rootValue"},
            None,
        )

    @mark.asyncio
    async def returns_an_awaitable_if_any_field_is_asynchronous():
        doc = "query Example { syncField, asyncField }"
        result = execute(schema, parse(doc), "rootValue")
        assert isawaitable(result)
        assert await result == (
            {"syncField": "rootValue", "asyncField": "rootValue"},
            None,
        )

    def describe_execute_sync():
        def does_not_return_an_awaitable_for_sync_execution():
            doc = "query Example { syncField }"
            result = execute_sync(schema, document=parse(doc), root_value="rootValue")
            assert result == (
                {"syncField": "rootValue"},
                None,
            )

        def does_not_throw_if_not_encountering_async_execution_with_check_sync():
            doc = "query Example { syncField }"
            result = execute_sync(
                schema, document=parse(doc), root_value="rootValue", check_sync=True
            )
            assert result == (
                {"syncField": "rootValue"},
                None,
            )

        @mark.asyncio
        @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
        async def throws_if_encountering_async_execution_with_check_sync():
            doc = "query Example { syncField, asyncField }"
            with raises(RuntimeError) as exc_info:
                execute_sync(
                    schema, document=parse(doc), root_value="rootValue", check_sync=True
                )
            msg = str(exc_info.value)
            assert msg == "GraphQL execution failed to complete synchronously."

        @mark.asyncio
        @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
        async def throws_if_encountering_async_operation_without_check_sync():
            doc = "query Example { syncField, asyncField }"
            result = execute_sync(schema, document=parse(doc), root_value="rootValue")
            assert result == (
                {"syncField": "rootValue", "asyncField": None},
                [
                    {
                        "message": "String cannot represent value:"
                        " <coroutine _resolve_async>",
                        "locations": [(1, 28)],
                        "path": ["asyncField"],
                    }
                ],
            )
            # garbage collect coroutine in order to not postpone the warning
            del result
            collect()

    def describe_graphql_sync():
        def reports_errors_raised_during_schema_validation():
            bad_schema = GraphQLSchema()
            result = graphql_sync(schema=bad_schema, source="{ __typename }")
            assert result == (None, [{"message": "Query root type must be provided."}])

        def does_not_return_an_awaitable_for_syntax_errors():
            doc = "fragment Example on Query { { { syncField }"
            assert graphql_sync(schema, doc) == (
                None,
                [
                    {
                        "message": "Syntax Error: Expected Name, found '{'.",
                        "locations": [(1, 29)],
                    }
                ],
            )

        def does_not_return_an_awaitable_for_validation_errors():
            doc = "fragment Example on Query { unknownField }"
            validation_errors = validate(schema, parse(doc))
            result = graphql_sync(schema, doc)
            assert result == (None, validation_errors)

        def does_not_return_an_awaitable_for_sync_execution():
            doc = "query Example { syncField }"
            assert graphql_sync(schema, doc, "rootValue") == (
                {"syncField": "rootValue"},
                None,
            )

        def does_not_throw_if_not_encountering_async_operation_with_check_sync():
            doc = "query Example { syncField }"
            assert graphql_sync(schema, doc, "rootValue") == (
                {"syncField": "rootValue"},
                None,
            )

        @mark.asyncio
        @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
        async def throws_if_encountering_async_operation_with_check_sync():
            doc = "query Example { syncField, asyncField }"
            with raises(RuntimeError) as exc_info:
                graphql_sync(schema, doc, "rootValue", check_sync=True)
            msg = str(exc_info.value)
            assert msg == "GraphQL execution failed to complete synchronously."

        @mark.asyncio
        @mark.filterwarnings("ignore:.* was never awaited:RuntimeWarning")
        async def throws_if_encountering_async_operation_without_check_sync():
            doc = "query Example { syncField, asyncField }"
            result = graphql_sync(schema, doc, "rootValue")
            assert result == (
                {"syncField": "rootValue", "asyncField": None},
                [
                    {
                        "message": "String cannot represent value:"
                        " <coroutine _resolve_async>",
                        "locations": [(1, 28)],
                        "path": ["asyncField"],
                    }
                ],
            )
            # garbage collect coroutine in order to not postpone the warning
            del result
            collect()



"""GraphQL Subscription

The :mod:`graphql.subscription` package is responsible for subscribing to updates
on specific data.

.. deprecated:: 3.2
   This package has been deprecated with its exported functions integrated into the
   :mod:`graphql.execution` package, to better conform with the terminology of the
   GraphQL specification. For backwards compatibility, the :mod:`graphql.subscription`
   package currently re-exports the moved functions from the :mod:`graphql.execution`
   package. In v3.3, the :mod:`graphql.subscription` package will be dropped entirely.
"""

from ..execution import subscribe, create_source_event_stream, MapAsyncIterator

__all__ = ["subscribe", "create_source_event_stream", "MapAsyncIterator"]



from graphql.language import parse, FieldNode, NameNode, OperationType, SelectionSetNode
from graphql.utilities import ast_to_dict


def describe_ast_to_disc():
    def converts_name_node_to_dict():
        node = NameNode(value="test")
        res = ast_to_dict(node)
        assert res == {"kind": "name", "value": "test"}
        assert list(res)[0] == "kind"
        assert ast_to_dict(node, locations=True) == res
        assert node.to_dict() == res
        assert node.to_dict(locations=True) == res

    def converts_two_name_nodes_to_list():
        nodes = [NameNode(value="foo"), NameNode(value="bar")]
        res = ast_to_dict(nodes)
        assert ast_to_dict(nodes, locations=True) == res
        assert res == [
            {"kind": "name", "value": "foo"},
            {"kind": "name", "value": "bar"},
        ]

    def converts_operation_type_to_its_value():
        assert ast_to_dict(OperationType.QUERY) == "query"

    def keeps_all_other_leaf_nodes():
        assert ast_to_dict(None) is None  # type: ignore
        assert ast_to_dict(42) == 42  # type: ignore
        assert ast_to_dict("foo") == "foo"  # type: ignore
        ast = {"foo": "bar"}
        assert ast_to_dict(ast) is ast  # type: ignore

    def converts_recursive_ast_to_recursive_dict():
        field = FieldNode(name="foo", arguments=(), selection_set=())
        ast = SelectionSetNode(selections=(field,))
        field.selection_set = ast
        res = ast_to_dict(ast)
        assert res == {
            "kind": "selection_set",
            "selections": [
                {
                    "kind": "field",
                    "name": "foo",
                    "alias": None,
                    "arguments": [],
                    "directives": None,
                    "selection_set": res,
                }
            ],
        }

    def converts_simple_schema_to_dict():
        ast = parse(
            """
            type Query {
              me: User
            }

            type User {
              id: ID
              name: String
            }
            """
        )
        res = ast_to_dict(ast)
        assert ast.to_dict() == res
        assert res == {
            "definitions": [
                {
                    "description": None,
                    "directives": [],
                    "fields": [
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "name": {"kind": "name", "value": "me"},
                            "type": {
                                "kind": "named_type",
                                "name": {"kind": "name", "value": "User"},
                            },
                        }
                    ],
                    "interfaces": [],
                    "kind": "object_type_definition",
                    "name": {"kind": "name", "value": "Query"},
                },
                {
                    "description": None,
                    "directives": [],
                    "fields": [
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "name": {"kind": "name", "value": "id"},
                            "type": {
                                "kind": "named_type",
                                "name": {"kind": "name", "value": "ID"},
                            },
                        },
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "name": {"kind": "name", "value": "name"},
                            "type": {
                                "kind": "named_type",
                                "name": {"kind": "name", "value": "String"},
                            },
                        },
                    ],
                    "interfaces": [],
                    "kind": "object_type_definition",
                    "name": {"kind": "name", "value": "User"},
                },
            ],
            "kind": "document",
        }
        assert list(res)[0] == "kind"

    def converts_simple_schema_to_dict_with_locations():
        ast = parse(
            """
            type Query {
              me: User
            }

            type User {
              id: ID
              name: String
            }
            """
        )
        res = ast_to_dict(ast, locations=True)
        assert ast.to_dict(locations=True) == res
        assert res == {
            "definitions": [
                {
                    "description": None,
                    "directives": [],
                    "fields": [
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "loc": {"end": 48, "start": 40},
                            "name": {
                                "kind": "name",
                                "loc": {"end": 42, "start": 40},
                                "value": "me",
                            },
                            "type": {
                                "kind": "named_type",
                                "loc": {"end": 48, "start": 44},
                                "name": {
                                    "kind": "name",
                                    "loc": {"end": 48, "start": 44},
                                    "value": "User",
                                },
                            },
                        }
                    ],
                    "interfaces": [],
                    "kind": "object_type_definition",
                    "loc": {"end": 62, "start": 13},
                    "name": {
                        "kind": "name",
                        "loc": {"end": 23, "start": 18},
                        "value": "Query",
                    },
                },
                {
                    "description": None,
                    "directives": [],
                    "fields": [
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "loc": {"end": 108, "start": 102},
                            "name": {
                                "kind": "name",
                                "loc": {"end": 104, "start": 102},
                                "value": "id",
                            },
                            "type": {
                                "kind": "named_type",
                                "loc": {"end": 108, "start": 106},
                                "name": {
                                    "kind": "name",
                                    "loc": {"end": 108, "start": 106},
                                    "value": "ID",
                                },
                            },
                        },
                        {
                            "arguments": [],
                            "description": None,
                            "directives": [],
                            "kind": "field_definition",
                            "loc": {"end": 135, "start": 123},
                            "name": {
                                "kind": "name",
                                "loc": {"end": 127, "start": 123},
                                "value": "name",
                            },
                            "type": {
                                "kind": "named_type",
                                "loc": {"end": 135, "start": 129},
                                "name": {
                                    "kind": "name",
                                    "loc": {"end": 135, "start": 129},
                                    "value": "String",
                                },
                            },
                        },
                    ],
                    "interfaces": [],
                    "kind": "object_type_definition",
                    "loc": {"end": 149, "start": 76},
                    "name": {
                        "kind": "name",
                        "loc": {"end": 85, "start": 81},
                        "value": "User",
                    },
                },
            ],
            "kind": "document",
            "loc": {"end": 162, "start": 0},
        }
        keys = list(res)
        assert keys[0] == "kind"
        assert keys[-1] == "loc"
        assert list(res["loc"]) == ["start", "end"]

    def converts_simple_query_to_dict():
        ast = parse(
            """
            query HeroForEpisode($ep: Episode!) {
              hero(episode: $ep) {
                name
                ... on Droid {
                  primaryFunction
                }
                ... on Human {
                  height
                }
              }
            }
            """
        )
        res = ast_to_dict(ast)
        assert ast.to_dict() == res
        assert res == {
            "definitions": [
                {
                    "directives": [],
                    "kind": "operation_definition",
                    "name": {"kind": "name", "value": "HeroForEpisode"},
                    "operation": "query",
                    "selection_set": {
                        "kind": "selection_set",
                        "selections": [
                            {
                                "alias": None,
                                "arguments": [
                                    {
                                        "kind": "argument",
                                        "name": {"kind": "name", "value": "episode"},
                                        "value": {
                                            "kind": "variable",
                                            "name": {"kind": "name", "value": "ep"},
                                        },
                                    }
                                ],
                                "directives": [],
                                "kind": "field",
                                "name": {"kind": "name", "value": "hero"},
                                "selection_set": {
                                    "kind": "selection_set",
                                    "selections": [
                                        {
                                            "alias": None,
                                            "arguments": [],
                                            "directives": [],
                                            "kind": "field",
                                            "name": {"kind": "name", "value": "name"},
                                            "selection_set": None,
                                        },
                                        {
                                            "directives": [],
                                            "kind": "inline_fragment",
                                            "selection_set": {
                                                "kind": "selection_set",
                                                "selections": [
                                                    {
                                                        "alias": None,
                                                        "arguments": [],
                                                        "directives": [],
                                                        "kind": "field",
                                                        "name": {
                                                            "kind": "name",
                                                            "value": "primaryFunction",
                                                        },
                                                        "selection_set": None,
                                                    }
                                                ],
                                            },
                                            "type_condition": {
                                                "kind": "named_type",
                                                "name": {
                                                    "kind": "name",
                                                    "value": "Droid",
                                                },
                                            },
                                        },
                                        {
                                            "directives": [],
                                            "kind": "inline_fragment",
                                            "selection_set": {
                                                "kind": "selection_set",
                                                "selections": [
                                                    {
                                                        "alias": None,
                                                        "arguments": [],
                                                        "directives": [],
                                                        "kind": "field",
                                                        "name": {
                                                            "kind": "name",
                                                            "value": "height",
                                                        },
                                                        "selection_set": None,
                                                    }
                                                ],
                                            },
                                            "type_condition": {
                                                "kind": "named_type",
                                                "name": {
                                                    "kind": "name",
                                                    "value": "Human",
                                                },
                                            },
                                        },
                                    ],
                                },
                            }
                        ],
                    },
                    "variable_definitions": [
                        {
                            "default_value": None,
                            "directives": [],
                            "kind": "variable_definition",
                            "type": {
                                "kind": "non_null_type",
                                "type": {
                                    "kind": "named_type",
                                    "name": {"kind": "name", "value": "Episode"},
                                },
                            },
                            "variable": {
                                "kind": "variable",
                                "name": {"kind": "name", "value": "ep"},
                            },
                        }
                    ],
                }
            ],
            "kind": "document",
        }
        assert list(res)[0] == "kind"

    def converts_simple_query_to_dict_with_locations():
        ast = parse(
            """
            query HeroForEpisode($ep: Episode!) {
              hero(episode: $ep) {
                name
                ... on Droid {
                  primaryFunction
                }
                ... on Human {
                  height
                }
              }
            }
            """
        )
        res = ast_to_dict(ast, locations=True)
        assert ast.to_dict(locations=True) == res
        assert res == {
            "definitions": [
                {
                    "directives": [],
                    "kind": "operation_definition",
                    "loc": {"end": 293, "start": 13},
                    "name": {
                        "kind": "name",
                        "loc": {"end": 33, "start": 19},
                        "value": "HeroForEpisode",
                    },
                    "operation": "query",
                    "selection_set": {
                        "kind": "selection_set",
                        "loc": {"end": 293, "start": 49},
                        "selections": [
                            {
                                "alias": None,
                                "arguments": [
                                    {
                                        "kind": "argument",
                                        "loc": {"end": 82, "start": 70},
                                        "name": {
                                            "kind": "name",
                                            "loc": {"end": 77, "start": 70},
                                            "value": "episode",
                                        },
                                        "value": {
                                            "kind": "variable",
                                            "loc": {"end": 82, "start": 79},
                                            "name": {
                                                "kind": "name",
                                                "loc": {"end": 82, "start": 80},
                                                "value": "ep",
                                            },
                                        },
                                    }
                                ],
                                "directives": [],
                                "kind": "field",
                                "loc": {"end": 279, "start": 65},
                                "name": {
                                    "kind": "name",
                                    "loc": {"end": 69, "start": 65},
                                    "value": "hero",
                                },
                                "selection_set": {
                                    "kind": "selection_set",
                                    "loc": {"end": 279, "start": 84},
                                    "selections": [
                                        {
                                            "alias": None,
                                            "arguments": [],
                                            "directives": [],
                                            "kind": "field",
                                            "loc": {"end": 106, "start": 102},
                                            "name": {
                                                "kind": "name",
                                                "loc": {"end": 106, "start": 102},
                                                "value": "name",
                                            },
                                            "selection_set": None,
                                        },
                                        {
                                            "directives": [],
                                            "kind": "inline_fragment",
                                            "loc": {"end": 189, "start": 123},
                                            "selection_set": {
                                                "kind": "selection_set",
                                                "loc": {"end": 189, "start": 136},
                                                "selections": [
                                                    {
                                                        "alias": None,
                                                        "arguments": [],
                                                        "directives": [],
                                                        "kind": "field",
                                                        "loc": {
                                                            "end": 171,
                                                            "start": 156,
                                                        },
                                                        "name": {
                                                            "kind": "name",
                                                            "loc": {
                                                                "end": 171,
                                                                "start": 156,
                                                            },
                                                            "value": "primaryFunction",
                                                        },
                                                        "selection_set": None,
                                                    }
                                                ],
                                            },
                                            "type_condition": {
                                                "kind": "named_type",
                                                "loc": {"end": 135, "start": 130},
                                                "name": {
                                                    "kind": "name",
                                                    "loc": {"end": 135, "start": 130},
                                                    "value": "Droid",
                                                },
                                            },
                                        },
                                        {
                                            "directives": [],
                                            "kind": "inline_fragment",
                                            "loc": {"end": 263, "start": 206},
                                            "selection_set": {
                                                "kind": "selection_set",
                                                "loc": {"end": 263, "start": 219},
                                                "selections": [
                                                    {
                                                        "alias": None,
                                                        "arguments": [],
                                                        "directives": [],
                                                        "kind": "field",
                                                        "loc": {
                                                            "end": 245,
                                                            "start": 239,
                                                        },
                                                        "name": {
                                                            "kind": "name",
                                                            "loc": {
                                                                "end": 245,
                                                                "start": 239,
                                                            },
                                                            "value": "height",
                                                        },
                                                        "selection_set": None,
                                                    }
                                                ],
                                            },
                                            "type_condition": {
                                                "kind": "named_type",
                                                "loc": {"end": 218, "start": 213},
                                                "name": {
                                                    "kind": "name",
                                                    "loc": {"end": 218, "start": 213},
                                                    "value": "Human",
                                                },
                                            },
                                        },
                                    ],
                                },
                            }
                        ],
                    },
                    "variable_definitions": [
                        {
                            "default_value": None,
                            "directives": [],
                            "kind": "variable_definition",
                            "loc": {"end": 47, "start": 34},
                            "type": {
                                "kind": "non_null_type",
                                "loc": {"end": 47, "start": 39},
                                "type": {
                                    "kind": "named_type",
                                    "loc": {"end": 46, "start": 39},
                                    "name": {
                                        "kind": "name",
                                        "loc": {"end": 46, "start": 39},
                                        "value": "Episode",
                                    },
                                },
                            },
                            "variable": {
                                "kind": "variable",
                                "loc": {"end": 37, "start": 34},
                                "name": {
                                    "kind": "name",
                                    "loc": {"end": 37, "start": 35},
                                    "value": "ep",
                                },
                            },
                        }
                    ],
                }
            ],
            "kind": "document",
            "loc": {"end": 306, "start": 0},
        }
        keys = list(res)
        assert keys[0] == "kind"
        assert keys[-1] == "loc"
        assert list(res["loc"]) == ["start", "end"]



"""User registry demo.

This is an additional end-to-end test and demo for running the basic GraphQL
operations on a simulated user registry database backend.
"""

from asyncio import sleep, wait
from collections import defaultdict
from enum import Enum
from inspect import isawaitable
from typing import Any, Dict, List, NamedTuple, Optional

try:
    from asyncio import create_task
except ImportError:  # Python < 3.7
    create_task = None  # type: ignore

from pytest import fixture, mark

from graphql import (
    graphql,
    parse,
    subscribe,
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLField,
    GraphQLID,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

from graphql.pyutils import SimplePubSub, SimplePubSubIterator
from graphql.execution.map_async_iterator import MapAsyncIterator


class User(NamedTuple):
    """A simple user object class."""

    firstName: str
    lastName: str
    tweets: Optional[int]
    id: Optional[str] = None
    verified: bool = False


class MutationEnum(Enum):
    """Mutation event type"""

    CREATED = "created"
    UPDATED = "updated"
    DELETED = "deleted"


class UserRegistry:
    """Simulation of a user registry with asynchronous database backend access."""

    def __init__(self, **users):
        self._registry: Dict[str, User] = users
        self._pubsub = defaultdict(SimplePubSub)

    async def get(self, id_: str) -> Optional[User]:
        """Get a user object from the registry"""
        await sleep(0)
        return self._registry.get(id_)

    async def create(self, **kwargs) -> User:
        """Create a user object in the registry"""
        await sleep(0)
        id_ = str(len(self._registry))
        user = User(id=id_, **kwargs)
        self._registry[id_] = user
        self.emit_event(MutationEnum.CREATED, user)
        return user

    async def update(self, id_: str, **kwargs) -> User:
        """Update a user object in the registry"""
        await sleep(0)
        # noinspection PyProtectedMember
        user = self._registry[id_]._replace(**kwargs)
        self._registry[id_] = user
        self.emit_event(MutationEnum.UPDATED, user)
        return user

    async def delete(self, id_: str) -> User:
        """Delete a user object in the registry"""
        await sleep(0)
        user = self._registry.pop(id_)
        self.emit_event(MutationEnum.DELETED, user)
        return user

    def emit_event(self, mutation: MutationEnum, user: User) -> None:
        """Emit mutation events for the given object and its class"""
        payload = {"user": user, "mutation": mutation.value}
        self._pubsub[None].emit(payload)  # notify all user subscriptions
        self._pubsub[user.id].emit(payload)  # notify single user subscriptions

    def event_iterator(self, id_: Optional[str]) -> SimplePubSubIterator:
        return self._pubsub[id_].get_subscriber()


mutation_type = GraphQLEnumType("MutationType", MutationEnum)

user_type = GraphQLObjectType(
    "UserType",
    {
        "id": GraphQLField(GraphQLNonNull(GraphQLID)),
        "firstName": GraphQLField(GraphQLNonNull(GraphQLString)),
        "lastName": GraphQLField(GraphQLNonNull(GraphQLString)),
        "tweets": GraphQLField(GraphQLInt),
        "verified": GraphQLField(GraphQLNonNull(GraphQLBoolean)),
    },
)

user_input_type = GraphQLInputObjectType(
    "UserInputType",
    {
        "firstName": GraphQLInputField(GraphQLNonNull(GraphQLString)),
        "lastName": GraphQLInputField(GraphQLNonNull(GraphQLString)),
        "tweets": GraphQLInputField(GraphQLInt),
        "verified": GraphQLInputField(GraphQLBoolean),
    },
)

subscription_user_type = GraphQLObjectType(
    "SubscriptionUserType",
    {"mutation": GraphQLField(mutation_type), "user": GraphQLField(user_type)},
)


async def resolve_user(_root, info, **args):
    """Resolver function for fetching a user object"""
    return await info.context["registry"].get(args["id"])


async def resolve_create_user(_root, info, data):
    """Resolver function for creating a user object"""
    user = await info.context["registry"].create(**data)
    return user


# noinspection PyShadowingBuiltins
async def resolve_update_user(_root, info, id, data):
    """Resolver function for updating a user object"""
    user = await info.context["registry"].update(id, **data)
    return user


# noinspection PyShadowingBuiltins
async def resolve_delete_user(_root, info, id):
    """Resolver function for deleting a user object"""
    user = await info.context["registry"].get(id)
    await info.context["registry"].delete(user.id)
    return True


# noinspection PyShadowingBuiltins
async def subscribe_user(_root, info, id=None):
    """Subscribe to mutations of a specific user object or all user objects"""
    async_iterator = info.context["registry"].event_iterator(id)
    async for event in async_iterator:
        yield await event if isawaitable(event) else event  # pragma: no cover exit


# noinspection PyShadowingBuiltins,PyUnusedLocal
async def resolve_subscription_user(event, info, id):
    """Resolver function for user subscriptions"""
    user = event["user"]
    mutation = MutationEnum(event["mutation"]).value
    return {"user": user, "mutation": mutation}


schema = GraphQLSchema(
    query=GraphQLObjectType(
        "RootQueryType",
        {
            "User": GraphQLField(
                user_type, args={"id": GraphQLArgument(GraphQLID)}, resolve=resolve_user
            )
        },
    ),
    mutation=GraphQLObjectType(
        "RootMutationType",
        {
            "createUser": GraphQLField(
                user_type,
                args={"data": GraphQLArgument(GraphQLNonNull(user_input_type))},
                resolve=resolve_create_user,
            ),
            "deleteUser": GraphQLField(
                GraphQLBoolean,
                args={"id": GraphQLArgument(GraphQLNonNull(GraphQLID))},
                resolve=resolve_delete_user,
            ),
            "updateUser": GraphQLField(
                user_type,
                args={
                    "id": GraphQLArgument(GraphQLNonNull(GraphQLID)),
                    "data": GraphQLArgument(GraphQLNonNull(user_input_type)),
                },
                resolve=resolve_update_user,
            ),
        },
    ),
    subscription=GraphQLObjectType(
        "RootSubscriptionType",
        {
            "subscribeUser": GraphQLField(
                subscription_user_type,
                args={"id": GraphQLArgument(GraphQLID)},
                subscribe=subscribe_user,
                resolve=resolve_subscription_user,
            )
        },
    ),
)


@fixture
def context():
    return {"registry": UserRegistry()}


def describe_query():
    @mark.asyncio
    async def query_user(context):
        user = await context["registry"].create(
            firstName="John", lastName="Doe", tweets=42, verified=True
        )

        query = """
            query ($userId: ID!) {
                User(id: $userId) {
                    id, firstName, lastName, tweets, verified
                }
            }
            """

        variables = {"userId": user.id}
        result = await graphql(
            schema, query, context_value=context, variable_values=variables
        )

        assert not result.errors
        assert result.data == {
            "User": {
                "id": user.id,
                "firstName": user.firstName,
                "lastName": user.lastName,
                "tweets": user.tweets,
                "verified": user.verified,
            }
        }


def describe_mutation():
    @mark.asyncio
    async def create_user(context):
        received = {}

        def subscriber(event_name):
            def receive(msg):
                received[event_name] = msg

            return receive

        # noinspection PyProtectedMember
        pubsub = context["registry"]._pubsub
        pubsub[None].subscribers.add(subscriber("User"))
        pubsub["0"].subscribers.add(subscriber("User 0"))

        query = """
            mutation ($userData: UserInputType!) {
                createUser(data: $userData) {
                    id, firstName, lastName, tweets, verified
                }
            }
            """
        user_data = dict(firstName="John", lastName="Doe", tweets=42, verified=True)
        variables = {"userData": user_data}
        result = await graphql(
            schema, query, context_value=context, variable_values=variables
        )

        user = await context["registry"].get("0")
        assert user == User(id="0", **user_data)  # type: ignore

        assert result.errors is None
        assert result.data == {
            "createUser": {
                "id": user.id,
                "firstName": user.firstName,
                "lastName": user.lastName,
                "tweets": user.tweets,
                "verified": user.verified,
            }
        }

        assert received == {
            "User": {"user": user, "mutation": MutationEnum.CREATED.value},
            "User 0": {"user": user, "mutation": MutationEnum.CREATED.value},
        }

    @mark.asyncio
    async def update_user(context):
        received = {}

        def subscriber(event_name):
            def receive(msg):
                received[event_name] = msg

            return receive

        # noinspection PyProtectedMember
        pubsub = context["registry"]._pubsub
        pubsub[None].subscribers.add(subscriber("User"))
        pubsub["0"].subscribers.add(subscriber("User 0"))

        user = await context["registry"].create(
            firstName="John", lastName="Doe", tweets=42, verified=True
        )
        user_data = {
            "firstName": "Jane",
            "lastName": "Roe",
            "tweets": 210,
            "verified": False,
        }

        query = """
            mutation ($userId: ID!, $userData: UserInputType!) {
                updateUser(id: $userId, data: $userData) {
                    id, firstName, lastName, tweets, verified
                }
            }"""

        variables = {"userId": user.id, "userData": user_data}
        result = await graphql(
            schema, query, context_value=context, variable_values=variables
        )

        user = await context["registry"].get("0")
        assert user == User(id="0", **user_data)  # type: ignore

        assert result.errors is None
        assert result.data == {
            "updateUser": {
                "id": user.id,
                "firstName": user.firstName,
                "lastName": user.lastName,
                "tweets": user.tweets,
                "verified": user.verified,
            }
        }

        assert received == {
            "User": {"user": user, "mutation": MutationEnum.UPDATED.value},
            "User 0": {"user": user, "mutation": MutationEnum.UPDATED.value},
        }

    @mark.asyncio
    async def delete_user(context):
        received = {}

        def subscriber(name):
            def receive(msg):
                received[name] = msg

            return receive

        # noinspection PyProtectedMember
        pubsub = context["registry"]._pubsub
        pubsub[None].subscribers.add(subscriber("User"))
        pubsub["0"].subscribers.add(subscriber("User 0"))

        user = await context["registry"].create(
            firstName="John", lastName="Doe", tweets=42, verified=True
        )

        query = """
            mutation ($userId: ID!) {
                deleteUser(id: $userId)
            }
            """

        variables = {"userId": user.id}
        result = await graphql(
            schema, query, context_value=context, variable_values=variables
        )

        assert result.errors is None
        assert result.data == {"deleteUser": True}

        assert await context["registry"].get(user.id) is None

        assert received == {
            "User": {"user": user, "mutation": MutationEnum.DELETED.value},
            "User 0": {"user": user, "mutation": MutationEnum.DELETED.value},
        }


def describe_subscription():
    @mark.asyncio
    async def subscribe_to_user_mutations(context):
        query = """
            subscription ($userId: ID!) {
                subscribeUser(id: $userId) {
                    mutation
                    user { id, firstName, lastName, tweets, verified }
                }
            }
            """

        variables = {"userId": "0"}
        subscription_one = await subscribe(
            schema, parse(query), context_value=context, variable_values=variables
        )
        assert isinstance(subscription_one, MapAsyncIterator)

        query = """
            subscription {
                subscribeUser(id: null) {
                    mutation
                    user { id, firstName, lastName, tweets, verified }
                }
            }
            """

        subscription_all = await subscribe(schema, parse(query), context_value=context)
        assert isinstance(subscription_all, MapAsyncIterator)

        received_one = []
        received_all = []

        async def mutate_users():
            await sleep(0)  # make sure subscribers are running
            await graphql(
                schema,
                """
                mutation {createUser(data: {
                    firstName: "John"
                    lastName: "Doe"
                    tweets: 42
                    verified: true}) { id }
                }""",
                context_value=context,
            )
            await graphql(
                schema,
                """
                mutation {createUser(data: {
                    firstName: "James"
                    lastName: "Doe"
                    tweets: 4
                    verified: false}) { id }
                }""",
                context_value=context,
            )
            await graphql(
                schema,
                """
                mutation {updateUser(id: 0, data: {
                    firstName: "Jane"
                    lastName: "Roe"
                    tweets: 210
                    verified: false}) { id }
                }""",
                context_value=context,
            )
            await graphql(
                schema,
                """
                mutation {updateUser(id: 1, data: {
                    firstName: "Janette"
                    lastName: "Roe"
                    tweets: 20
                    verified: true}) { id }
                }""",
                context_value=context,
            )
            await graphql(
                schema,
                """
                mutation {deleteUser(id: "0")}
                """,
                context_value=context,
            )
            await graphql(
                schema,
                """
                mutation {deleteUser(id: "1")}
                """,
                context_value=context,
            )

        async def receive_one():
            async for result in subscription_one:  # type: ignore # pragma: no cover
                received_one.append(result)
                if len(received_one) == 3:  # pragma: no cover else
                    break

        async def receive_all():
            async for result in subscription_all:  # type: ignore # pragma: no cover
                received_all.append(result)
                if len(received_all) == 6:  # pragma: no cover else
                    break

        tasks = [
            create_task(task()) if create_task else task()
            for task in (mutate_users, receive_one, receive_all)
        ]
        done, pending = await wait(tasks, timeout=1)
        assert not pending

        expected_data: List[Dict[str, Any]] = [
            {
                "mutation": "CREATED",
                "user": {
                    "id": "0",
                    "firstName": "John",
                    "lastName": "Doe",
                    "tweets": 42,
                    "verified": True,
                },
            },
            {
                "mutation": "CREATED",
                "user": {
                    "id": "1",
                    "firstName": "James",
                    "lastName": "Doe",
                    "tweets": 4,
                    "verified": False,
                },
            },
            {
                "mutation": "UPDATED",
                "user": {
                    "id": "0",
                    "firstName": "Jane",
                    "lastName": "Roe",
                    "tweets": 210,
                    "verified": False,
                },
            },
            {
                "mutation": "UPDATED",
                "user": {
                    "id": "1",
                    "firstName": "Janette",
                    "lastName": "Roe",
                    "tweets": 20,
                    "verified": True,
                },
            },
            {
                "mutation": "DELETED",
                "user": {
                    "id": "0",
                    "firstName": "Jane",
                    "lastName": "Roe",
                    "tweets": 210,
                    "verified": False,
                },
            },
            {
                "mutation": "DELETED",
                "user": {
                    "id": "1",
                    "firstName": "Janette",
                    "lastName": "Roe",
                    "tweets": 20,
                    "verified": True,
                },
            },
        ]

        assert received_one == [
            ({"subscribeUser": data}, None)
            for data in expected_data
            if data["user"]["id"] == "0"
        ]
        assert received_all == [
            ({"subscribeUser": data}, None) for data in expected_data
        ]

        await sleep(0)



from typing import Any

from ...error import GraphQLError
from ...language import (
    FragmentDefinitionNode,
    InlineFragmentNode,
    print_ast,
)
from ...type import is_composite_type
from ...utilities import type_from_ast
from . import ValidationRule

__all__ = ["FragmentsOnCompositeTypesRule"]


class FragmentsOnCompositeTypesRule(ValidationRule):
    """Fragments on composite type

    Fragments use a type condition to determine if they apply, since fragments can only
    be spread into a composite type (object, interface, or union), the type condition
    must also be a composite type.

    See https://spec.graphql.org/draft/#sec-Fragments-On-Composite-Types
    """

    def enter_inline_fragment(self, node: InlineFragmentNode, *_args: Any) -> None:
        type_condition = node.type_condition
        if type_condition:
            type_ = type_from_ast(self.context.schema, type_condition)
            if type_ and not is_composite_type(type_):
                type_str = print_ast(type_condition)
                self.report_error(
                    GraphQLError(
                        "Fragment cannot condition"
                        f" on non composite type '{type_str}'.",
                        type_condition,
                    )
                )

    def enter_fragment_definition(
        self, node: FragmentDefinitionNode, *_args: Any
    ) -> None:
        type_condition = node.type_condition
        type_ = type_from_ast(self.context.schema, type_condition)
        if type_ and not is_composite_type(type_):
            type_str = print_ast(type_condition)
            self.report_error(
                GraphQLError(
                    f"Fragment '{node.name.value}' cannot condition"
                    f" on non composite type '{type_str}'.",
                    type_condition,
                )
            )



from typing import cast, List, Union

from pytest import raises

from graphql.error import GraphQLError
from graphql.language import (
    parse,
    Node,
    OperationDefinitionNode,
    ObjectTypeDefinitionNode,
    Source,
)

from ..utils import dedent


source = Source(
    dedent(
        """
        {
          field
        }
        """
    )
)

ast = parse(source)
operation_node = ast.definitions[0]
operation_node = cast(OperationDefinitionNode, operation_node)
assert operation_node and operation_node.kind == "operation_definition"
field_node = operation_node.selection_set.selections[0]
assert field_node


def describe_graphql_error():
    def is_a_class_and_is_a_subclass_of_exception():
        assert type(GraphQLError) is type
        assert issubclass(GraphQLError, Exception)
        assert isinstance(GraphQLError("str"), Exception)
        assert isinstance(GraphQLError("str"), GraphQLError)

    def has_a_name_message_extensions_and_stack_trace():
        e = GraphQLError("msg")
        assert e.__class__.__name__ == "GraphQLError"
        assert e.message == "msg"
        assert e.extensions == {}
        assert e.__traceback__ is None
        assert str(e) == "msg"

    def can_pass_positional_and_keyword_arguments():
        e1 = GraphQLError(
            "msg",
            [field_node],
            source,
            [1, 2, 3],
            ["a", "b", "c"],
            Exception("test"),
            {"foo": "bar"},
        )
        e2 = GraphQLError(
            message="msg",
            nodes=[field_node],
            source=source,
            positions=[1, 2, 3],
            path=["a", "b", "c"],
            original_error=Exception("test"),
            extensions={"foo": "bar"},
        )
        assert e1 == e2

    def formatted_dict_has_only_keys_prescribed_in_the_spec():
        e = GraphQLError(
            "msg",
            [field_node],
            source,
            [1, 2, 3],
            ["a", "b", "c"],
            Exception("test"),
            {"foo": "bar"},
        )
        assert set(e.formatted) == {"message", "path", "locations", "extensions"}

    def uses_the_stack_of_an_original_error():
        try:
            raise RuntimeError("original")
        except RuntimeError as runtime_error:
            original = runtime_error
        e = GraphQLError("msg", original_error=original)
        assert e.__class__.__name__ == "GraphQLError"
        assert e.__traceback__ is original.__traceback__
        assert e.message == "msg"
        assert e.original_error is original
        assert str(e.original_error) == "original"

    def passes_the_context_of_an_original_error():
        context = ValueError("cause")
        try:
            raise context
        except ValueError:
            try:
                raise RuntimeError("effect")
            except RuntimeError as runtime_error:
                original = runtime_error
        e = GraphQLError("msg", original_error=original)
        assert e.__context__ is context

    def passes_the_cause_of_an_original_error():
        cause = ValueError("cause")
        try:
            raise RuntimeError("effect") from cause
        except RuntimeError as runtime_error:
            original = runtime_error
        e = GraphQLError("msg", original_error=original)
        assert e.__cause__ is cause

    def creates_new_stack_if_original_error_has_no_stack():
        try:
            raise RuntimeError
        except RuntimeError as original_with_traceback:
            original_traceback = original_with_traceback.__traceback__
            original = RuntimeError("original")
            e = GraphQLError("msg", original_error=original)
        assert e.__class__.__name__ == "GraphQLError"
        assert original.__traceback__ is None
        assert original_traceback is not None
        assert e.__traceback__ is original_traceback
        assert e.message == "msg"
        assert e.original_error is original
        assert str(e.original_error) == "original"

    def converts_nodes_to_positions_and_locations():
        e = GraphQLError("msg", [field_node])
        assert e.nodes == [field_node]
        assert e.source is source
        assert e.positions == [4]
        assert e.locations == [(2, 3)]

    def converts_single_node_to_positions_and_locations():
        e = GraphQLError("msg", field_node)  # Non-array value.
        assert e.nodes == [field_node]
        assert e.source is source
        assert e.positions == [4]
        assert e.locations == [(2, 3)]

    def converts_node_with_loc_start_zero_to_positions_and_locations():
        e = GraphQLError("msg", operation_node)
        assert e.nodes == [operation_node]
        assert e.source is source
        assert e.positions == [0]
        assert e.locations == [(1, 1)]

    def converts_node_without_location_to_source_positions_and_locations_as_none():
        document_node = parse("{ foo }", no_location=True)

        e = GraphQLError("msg", document_node)
        assert e.nodes == [document_node]
        assert e.source is None
        assert e.positions is None
        assert e.locations is None

    def converts_source_and_positions_to_locations():
        e = GraphQLError("msg", None, source, [6])
        assert e.nodes is None
        assert e.source is source
        assert e.positions == [6]
        assert e.locations == [(2, 5)]

    def defaults_to_original_error_extension_only_if_arg_is_not_passed():
        original_extensions = {"original": "extensions"}
        original_error = GraphQLError("original", extensions=original_extensions)
        inherited_error = GraphQLError("InheritedError", original_error=original_error)
        assert inherited_error.message == "InheritedError"
        assert inherited_error.original_error is original_error
        assert inherited_error.extensions is original_extensions

        own_extensions = {"own": "extensions"}
        own_error = GraphQLError(
            "OwnError", original_error=original_error, extensions=own_extensions
        )
        assert own_error.message == "OwnError"
        assert own_error.original_error is original_error
        assert own_error.extensions is own_extensions

        own_empty_error = GraphQLError(
            "OwnEmptyError", original_error=original_error, extensions={}
        )
        assert own_empty_error.message == "OwnEmptyError"
        assert own_empty_error.original_error is original_error
        assert own_empty_error.extensions == {}

    def serializes_to_include_message():
        e = GraphQLError("msg")
        assert str(e) == "msg"
        assert repr(e) == "GraphQLError('msg')"

    def serializes_to_include_message_and_locations():
        e = GraphQLError("msg", field_node)
        assert "msg" in str(e)
        assert ":2:3" in str(e)
        assert repr(e) == (
            "GraphQLError('msg', locations=[SourceLocation(line=2, column=3)])"
        )
        assert e.formatted == {
            "locations": [{"column": 3, "line": 2}],
            "message": "msg",
        }

    def serializes_to_include_path():
        path: List[Union[int, str]] = ["path", 3, "to", "field"]
        e = GraphQLError("msg", path=path)
        assert e.path is path
        assert repr(e) == "GraphQLError('msg', path=['path', 3, 'to', 'field'])"
        assert e.formatted == {
            "message": "msg",
            "path": ["path", 3, "to", "field"],
        }

    def serializes_to_include_all_standard_fields():
        e_short = GraphQLError("msg")
        assert str(e_short) == "msg"
        assert repr(e_short) == "GraphQLError('msg')"

        path: List[Union[str, int]] = ["path", 2, "field"]
        extensions = {"foo": "bar "}
        e_full = GraphQLError("msg", field_node, None, None, path, None, extensions)
        assert str(e_full) == (
            "msg\n\nGraphQL request:2:3\n" "1 | {\n2 |   field\n  |   ^\n3 | }"
        )
        assert repr(e_full) == (
            "GraphQLError('msg', locations=[SourceLocation(line=2, column=3)],"
            " path=['path', 2, 'field'], extensions={'foo': 'bar '})"
        )
        assert e_full.formatted == {
            "message": "msg",
            "locations": [{"line": 2, "column": 3}],
            "path": ["path", 2, "field"],
            "extensions": {"foo": "bar "},
        }

    def repr_includes_extensions():
        e = GraphQLError("msg", extensions={"foo": "bar"})
        assert repr(e) == "GraphQLError('msg', extensions={'foo': 'bar'})"

    def always_stores_path_as_list():
        path: List[Union[int, str]] = ["path", 3, "to", "field"]
        e = GraphQLError("msg,", path=tuple(path))
        assert isinstance(e.path, list)
        assert e.path == path

    def is_comparable():
        e1 = GraphQLError("msg,", path=["field", 1])
        assert e1 == e1
        assert e1 == e1.formatted
        assert not e1 != e1
        assert not e1 != e1.formatted
        e2 = GraphQLError("msg,", path=["field", 1])
        assert e1 == e2
        assert not e1 != e2
        assert e2.path and e2.path[1] == 1
        e2.path[1] = 2
        assert not e1 == e2
        assert e1 != e2
        assert not e1 == e2.formatted
        assert e1 != e2.formatted

    def is_hashable():
        hash(GraphQLError("msg"))

    def hashes_are_unique_per_instance():
        e1 = GraphQLError("msg")
        e2 = GraphQLError("msg")
        assert hash(e1) != hash(e2)


def describe_to_string():
    def deprecated_prints_an_error_using_print_error():
        # noinspection PyProtectedMember
        from graphql.error.graphql_error import print_error

        error = GraphQLError("Error")
        assert print_error(error) == "Error"
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            print_error(Exception)  # type: ignore
        assert str(exc_info.value) == "Expected a GraphQLError."

    def prints_an_error_without_location():
        error = GraphQLError("Error without location")
        assert str(error) == "Error without location"

    def prints_an_error_using_node_without_location():
        error = GraphQLError(
            "Error attached to node without location",
            parse("{ foo }", no_location=True),
        )
        assert str(error) == "Error attached to node without location"

    def prints_an_error_with_nodes_from_different_sources():
        doc_a = parse(
            Source(
                dedent(
                    """
                    type Foo {
                      field: String
                    }
                    """
                ),
                "SourceA",
            )
        )
        op_a = doc_a.definitions[0]
        op_a = cast(ObjectTypeDefinitionNode, op_a)
        assert op_a and op_a.kind == "object_type_definition" and op_a.fields
        field_a = op_a.fields[0]
        doc_b = parse(
            Source(
                dedent(
                    """
                    type Foo {
                      field: Int
                    }
                    """
                ),
                "SourceB",
            )
        )
        op_b = doc_b.definitions[0]
        op_b = cast(ObjectTypeDefinitionNode, op_b)
        assert op_b and op_b.kind == "object_type_definition" and op_b.fields
        field_b = op_b.fields[0]

        error = GraphQLError(
            "Example error with two nodes", [field_a.type, field_b.type]
        )

        assert str(error) == dedent(
            """
            Example error with two nodes

            SourceA:2:10
            1 | type Foo {
            2 |   field: String
              |          ^
            3 | }

            SourceB:2:10
            1 | type Foo {
            2 |   field: Int
              |          ^
            3 | }
            """
        )


def describe_formatted():
    def deprecated_formats_an_error_using_format_error():
        # noinspection PyProtectedMember
        from graphql.error.graphql_error import format_error

        error = GraphQLError("Example Error")
        assert format_error(error) == {
            "message": "Example Error",
        }
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            format_error(Exception)  # type: ignore
        assert str(exc_info.value) == "Expected a GraphQLError."

    def formats_graphql_error():
        path: List[Union[int, str]] = ["one", 2]
        extensions = {"ext": None}
        error = GraphQLError(
            "test message",
            Node(),
            Source(
                """
                query {
                  something
                }
                """
            ),
            [16, 41],
            ["one", 2],
            ValueError("original"),
            extensions=extensions,
        )
        assert error.formatted == {
            "message": "test message",
            "locations": [{"line": 2, "column": 16}, {"line": 3, "column": 17}],
            "path": path,
            "extensions": extensions,
        }

    def uses_default_message():
        # noinspection PyTypeChecker
        formatted = GraphQLError(None).formatted  # type: ignore

        assert formatted == {
            "message": "An unknown error occurred.",
        }

    def includes_path():
        path: List[Union[int, str]] = ["path", 3, "to", "field"]
        error = GraphQLError("msg", path=path)
        assert error.formatted == {"message": "msg", "path": path}

    def includes_extension_fields():
        error = GraphQLError("msg", extensions={"foo": "bar"})
        assert error.formatted == {
            "message": "msg",
            "extensions": {"foo": "bar"},
        }

    def can_be_created_from_dict():
        args = dict(
            nodes=[operation_node],
            source=source,
            positions=[6],
            path=["path", 2, "a"],
            original_error=Exception("I like turtles"),
            extensions=dict(hee="I like turtles"),
        )
        error = GraphQLError("msg", **args)  # type: ignore
        assert error.formatted == {
            "message": "msg",
            "locations": [{"column": 5, "line": 2}],
            "path": ["path", 2, "a"],
            "extensions": {"hee": "I like turtles"},
        }



from graphql import build_schema, parse, execute_sync
from graphql.utilities import get_introspection_query

from ..fixtures import big_schema_sdl  # noqa: F401


def test_execute_introspection_query(benchmark, big_schema_sdl):  # noqa: F811
    schema = build_schema(big_schema_sdl, assume_valid=True)
    document = parse(get_introspection_query())
    result = benchmark(lambda: execute_sync(schema=schema, document=document))
    assert result.errors is None



from typing import Collection, Dict, Optional, Tuple, Union, cast

from ..language import DirectiveLocation
from ..pyutils import inspect, merge_kwargs, natural_comparison_key
from ..type import (
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLUnionType,
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_introspection_type,
    is_list_type,
    is_non_null_type,
    is_object_type,
    is_scalar_type,
    is_union_type,
)

__all__ = ["lexicographic_sort_schema"]


def lexicographic_sort_schema(schema: GraphQLSchema) -> GraphQLSchema:
    """Sort GraphQLSchema.

    This function returns a sorted copy of the given GraphQLSchema.
    """

    def replace_type(
        type_: Union[GraphQLList, GraphQLNonNull, GraphQLNamedType]
    ) -> Union[GraphQLList, GraphQLNonNull, GraphQLNamedType]:
        if is_list_type(type_):
            return GraphQLList(replace_type(cast(GraphQLList, type_).of_type))
        if is_non_null_type(type_):
            return GraphQLNonNull(replace_type(cast(GraphQLNonNull, type_).of_type))
        return replace_named_type(cast(GraphQLNamedType, type_))

    def replace_named_type(type_: GraphQLNamedType) -> GraphQLNamedType:
        return type_map[type_.name]

    def replace_maybe_type(
        maybe_type: Optional[GraphQLNamedType],
    ) -> Optional[GraphQLNamedType]:
        return maybe_type and replace_named_type(maybe_type)

    def sort_directive(directive: GraphQLDirective) -> GraphQLDirective:
        return GraphQLDirective(
            **merge_kwargs(
                directive.to_kwargs(),
                locations=sorted(directive.locations, key=sort_by_name_key),
                args=sort_args(directive.args),
            )
        )

    def sort_args(args_map: Dict[str, GraphQLArgument]) -> Dict[str, GraphQLArgument]:
        args = {}
        for name, arg in sorted(args_map.items()):
            args[name] = GraphQLArgument(
                **merge_kwargs(
                    arg.to_kwargs(),
                    type_=replace_type(cast(GraphQLNamedType, arg.type)),
                )
            )
        return args

    def sort_fields(fields_map: Dict[str, GraphQLField]) -> Dict[str, GraphQLField]:
        fields = {}
        for name, field in sorted(fields_map.items()):
            fields[name] = GraphQLField(
                **merge_kwargs(
                    field.to_kwargs(),
                    type_=replace_type(cast(GraphQLNamedType, field.type)),
                    args=sort_args(field.args),
                )
            )
        return fields

    def sort_input_fields(
        fields_map: Dict[str, GraphQLInputField]
    ) -> Dict[str, GraphQLInputField]:
        return {
            name: GraphQLInputField(
                cast(
                    GraphQLInputType, replace_type(cast(GraphQLNamedType, field.type))
                ),
                description=field.description,
                default_value=field.default_value,
                ast_node=field.ast_node,
            )
            for name, field in sorted(fields_map.items())
        }

    def sort_types(array: Collection[GraphQLNamedType]) -> Tuple[GraphQLNamedType, ...]:
        return tuple(
            replace_named_type(type_) for type_ in sorted(array, key=sort_by_name_key)
        )

    def sort_named_type(type_: GraphQLNamedType) -> GraphQLNamedType:
        if is_scalar_type(type_) or is_introspection_type(type_):
            return type_
        if is_object_type(type_):
            type_ = cast(GraphQLObjectType, type_)
            return GraphQLObjectType(
                **merge_kwargs(
                    type_.to_kwargs(),
                    interfaces=lambda: sort_types(type_.interfaces),
                    fields=lambda: sort_fields(type_.fields),
                )
            )
        if is_interface_type(type_):
            type_ = cast(GraphQLInterfaceType, type_)
            return GraphQLInterfaceType(
                **merge_kwargs(
                    type_.to_kwargs(),
                    interfaces=lambda: sort_types(type_.interfaces),
                    fields=lambda: sort_fields(type_.fields),
                )
            )
        if is_union_type(type_):
            type_ = cast(GraphQLUnionType, type_)
            return GraphQLUnionType(
                **merge_kwargs(type_.to_kwargs(), types=lambda: sort_types(type_.types))
            )
        if is_enum_type(type_):
            type_ = cast(GraphQLEnumType, type_)
            return GraphQLEnumType(
                **merge_kwargs(
                    type_.to_kwargs(),
                    values={
                        name: GraphQLEnumValue(
                            val.value,
                            description=val.description,
                            deprecation_reason=val.deprecation_reason,
                            ast_node=val.ast_node,
                        )
                        for name, val in sorted(type_.values.items())
                    },
                )
            )
        if is_input_object_type(type_):
            type_ = cast(GraphQLInputObjectType, type_)
            return GraphQLInputObjectType(
                **merge_kwargs(
                    type_.to_kwargs(),
                    fields=lambda: sort_input_fields(type_.fields),
                )
            )

        # Not reachable. All possible types have been considered.
        raise TypeError(f"Unexpected type: {inspect(type_)}.")

    type_map: Dict[str, GraphQLNamedType] = {
        type_.name: sort_named_type(type_)
        for type_ in sorted(schema.type_map.values(), key=sort_by_name_key)
    }

    return GraphQLSchema(
        types=type_map.values(),
        directives=[
            sort_directive(directive)
            for directive in sorted(schema.directives, key=sort_by_name_key)
        ],
        query=cast(Optional[GraphQLObjectType], replace_maybe_type(schema.query_type)),
        mutation=cast(
            Optional[GraphQLObjectType], replace_maybe_type(schema.mutation_type)
        ),
        subscription=cast(
            Optional[GraphQLObjectType], replace_maybe_type(schema.subscription_type)
        ),
        ast_node=schema.ast_node,
    )


def sort_by_name_key(
    type_: Union[GraphQLNamedType, GraphQLDirective, DirectiveLocation]
) -> Tuple:
    return natural_comparison_key(type_.name)



from graphql.language import (
    FieldNode,
    NameNode,
    Node,
    OperationDefinitionNode,
    SelectionSetNode,
    parse,
    parse_value,
    print_ast,
    visit,
    Visitor,
)
from graphql.type import GraphQLSchema, get_named_type, is_composite_type
from graphql.utilities import TypeInfo, TypeInfoVisitor, build_schema

from ..fixtures import kitchen_sink_query  # noqa: F401


test_schema = build_schema(
    """
    interface Pet {
      name: String
    }

    type Dog implements Pet {
      name: String
    }

    type Cat implements Pet {
      name: String
    }

    type Human {
      name: String
      pets: [Pet]
    }

    type Alien {
      name(surname: Boolean): String
    }

    type QueryRoot {
      human(id: ID): Human
      alien: Alien
    }

    schema {
      query: QueryRoot
    }
    """
)


def describe_type_info():
    schema = GraphQLSchema()

    def allow_all_methods_to_be_called_before_entering_any_mode():
        type_info = TypeInfo(schema)

        assert type_info.get_type() is None
        assert type_info.get_parent_type() is None
        assert type_info.get_input_type() is None
        assert type_info.get_parent_input_type() is None
        assert type_info.get_field_def() is None
        assert type_info.get_default_value() is None
        assert type_info.get_directive() is None
        assert type_info.get_argument() is None
        assert type_info.get_enum_value() is None


def describe_visit_with_type_info():
    def supports_different_operation_types():
        schema = build_schema(
            """
            schema {
              query: QueryRoot
              mutation: MutationRoot
              subscription: SubscriptionRoot
            }

            type QueryRoot {
              foo: String
            }

            type MutationRoot {
              bar: String
            }

            type SubscriptionRoot {
              baz: String
            }
            """
        )
        ast = parse(
            """
            query { foo }
            mutation { bar }
            subscription { baz }
            """
        )

        class TestVisitor(Visitor):
            def __init__(self):
                super().__init__()
                self.root_types = {}

            def enter_operation_definition(self, node: OperationDefinitionNode, *_args):
                self.root_types[node.operation.value] = str(type_info.get_type())

        type_info = TypeInfo(schema)
        test_visitor = TestVisitor()
        assert visit(ast, TypeInfoVisitor(type_info, test_visitor))

        assert test_visitor.root_types == {
            "query": "QueryRoot",
            "mutation": "MutationRoot",
            "subscription": "SubscriptionRoot",
        }

    def provide_exact_same_arguments_to_wrapped_visitor():
        ast = parse("{ human(id: 4) { name, pets { ... { name } }, unknown } }")

        class TestVisitor(Visitor):
            def __init__(self):
                super().__init__()
                self.args = []

            def enter(self, *args):
                self.args.append(("enter", *args))

            def leave(self, *args):
                self.args.append(("leave", *args))

        test_visitor = TestVisitor()
        visit(ast, test_visitor)

        type_info = TypeInfo(test_schema)
        wrapped_visitor = TestVisitor()
        visit(ast, TypeInfoVisitor(type_info, wrapped_visitor))

        assert test_visitor.args == wrapped_visitor.args

    def maintains_type_info_during_visit():
        visited = []

        type_info = TypeInfo(test_schema)

        ast = parse("{ human(id: 4) { name, pets { ... { name } }, unknown } }")

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                input_type = type_info.get_input_type()
                node = args[0]
                visited.append(
                    (
                        "enter",
                        node.kind,
                        node.value if node.kind == "name" else None,
                        str(parent_type) if parent_type else None,
                        str(type_) if type_ else None,
                        str(input_type) if input_type else None,
                    )
                )

            @staticmethod
            def leave(*args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                input_type = type_info.get_input_type()
                node = args[0]
                visited.append(
                    (
                        "leave",
                        node.kind,
                        node.value if node.kind == "name" else None,
                        str(parent_type) if parent_type else None,
                        str(type_) if type_ else None,
                        str(input_type) if input_type else None,
                    )
                )

        visit(ast, TypeInfoVisitor(type_info, TestVisitor()))

        assert visited == [
            ("enter", "document", None, None, None, None),
            ("enter", "operation_definition", None, None, "QueryRoot", None),
            ("enter", "selection_set", None, "QueryRoot", "QueryRoot", None),
            ("enter", "field", None, "QueryRoot", "Human", None),
            ("enter", "name", "human", "QueryRoot", "Human", None),
            ("leave", "name", "human", "QueryRoot", "Human", None),
            ("enter", "argument", None, "QueryRoot", "Human", "ID"),
            ("enter", "name", "id", "QueryRoot", "Human", "ID"),
            ("leave", "name", "id", "QueryRoot", "Human", "ID"),
            ("enter", "int_value", None, "QueryRoot", "Human", "ID"),
            ("leave", "int_value", None, "QueryRoot", "Human", "ID"),
            ("leave", "argument", None, "QueryRoot", "Human", "ID"),
            ("enter", "selection_set", None, "Human", "Human", None),
            ("enter", "field", None, "Human", "String", None),
            ("enter", "name", "name", "Human", "String", None),
            ("leave", "name", "name", "Human", "String", None),
            ("leave", "field", None, "Human", "String", None),
            ("enter", "field", None, "Human", "[Pet]", None),
            ("enter", "name", "pets", "Human", "[Pet]", None),
            ("leave", "name", "pets", "Human", "[Pet]", None),
            ("enter", "selection_set", None, "Pet", "[Pet]", None),
            ("enter", "inline_fragment", None, "Pet", "Pet", None),
            ("enter", "selection_set", None, "Pet", "Pet", None),
            ("enter", "field", None, "Pet", "String", None),
            ("enter", "name", "name", "Pet", "String", None),
            ("leave", "name", "name", "Pet", "String", None),
            ("leave", "field", None, "Pet", "String", None),
            ("leave", "selection_set", None, "Pet", "Pet", None),
            ("leave", "inline_fragment", None, "Pet", "Pet", None),
            ("leave", "selection_set", None, "Pet", "[Pet]", None),
            ("leave", "field", None, "Human", "[Pet]", None),
            ("enter", "field", None, "Human", None, None),
            ("enter", "name", "unknown", "Human", None, None),
            ("leave", "name", "unknown", "Human", None, None),
            ("leave", "field", None, "Human", None, None),
            ("leave", "selection_set", None, "Human", "Human", None),
            ("leave", "field", None, "QueryRoot", "Human", None),
            ("leave", "selection_set", None, "QueryRoot", "QueryRoot", None),
            ("leave", "operation_definition", None, None, "QueryRoot", None),
            ("leave", "document", None, None, None, None),
        ]

    def maintains_type_info_during_edit():
        visited = []
        type_info = TypeInfo(test_schema)

        ast = parse("{ human(id: 4) { name, pets }, alien }")

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                input_type = type_info.get_input_type()
                node = args[0]
                visited.append(
                    (
                        "enter",
                        node.kind,
                        node.value if node.kind == "name" else None,
                        str(parent_type) if parent_type else None,
                        str(type_) if type_ else None,
                        str(input_type) if input_type else None,
                    )
                )

                # Make a query valid by adding missing selection sets.
                if (
                    node.kind == "field"
                    and not node.selection_set
                    and is_composite_type(get_named_type(type_))
                ):
                    return FieldNode(
                        alias=node.alias,
                        name=node.name,
                        arguments=node.arguments,
                        directives=node.directives,
                        selection_set=SelectionSetNode(
                            selections=[FieldNode(name=NameNode(value="__typename"))]
                        ),
                    )

            @staticmethod
            def leave(*args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                input_type = type_info.get_input_type()
                node = args[0]
                visited.append(
                    (
                        "leave",
                        node.kind,
                        node.value if node.kind == "name" else None,
                        str(parent_type) if parent_type else None,
                        str(type_) if type_ else None,
                        str(input_type) if input_type else None,
                    )
                )

        edited_ast = visit(ast, TypeInfoVisitor(type_info, TestVisitor()))

        assert ast == parse("{ human(id: 4) { name, pets }, alien }")

        assert print_ast(edited_ast) == print_ast(
            parse(
                "{ human(id: 4) { name, pets { __typename } },"
                " alien { __typename } }"
            )
        )

        assert visited == [
            ("enter", "document", None, None, None, None),
            ("enter", "operation_definition", None, None, "QueryRoot", None),
            ("enter", "selection_set", None, "QueryRoot", "QueryRoot", None),
            ("enter", "field", None, "QueryRoot", "Human", None),
            ("enter", "name", "human", "QueryRoot", "Human", None),
            ("leave", "name", "human", "QueryRoot", "Human", None),
            ("enter", "argument", None, "QueryRoot", "Human", "ID"),
            ("enter", "name", "id", "QueryRoot", "Human", "ID"),
            ("leave", "name", "id", "QueryRoot", "Human", "ID"),
            ("enter", "int_value", None, "QueryRoot", "Human", "ID"),
            ("leave", "int_value", None, "QueryRoot", "Human", "ID"),
            ("leave", "argument", None, "QueryRoot", "Human", "ID"),
            ("enter", "selection_set", None, "Human", "Human", None),
            ("enter", "field", None, "Human", "String", None),
            ("enter", "name", "name", "Human", "String", None),
            ("leave", "name", "name", "Human", "String", None),
            ("leave", "field", None, "Human", "String", None),
            ("enter", "field", None, "Human", "[Pet]", None),
            ("enter", "name", "pets", "Human", "[Pet]", None),
            ("leave", "name", "pets", "Human", "[Pet]", None),
            ("enter", "selection_set", None, "Pet", "[Pet]", None),
            ("enter", "field", None, "Pet", "String!", None),
            ("enter", "name", "__typename", "Pet", "String!", None),
            ("leave", "name", "__typename", "Pet", "String!", None),
            ("leave", "field", None, "Pet", "String!", None),
            ("leave", "selection_set", None, "Pet", "[Pet]", None),
            ("leave", "field", None, "Human", "[Pet]", None),
            ("leave", "selection_set", None, "Human", "Human", None),
            ("leave", "field", None, "QueryRoot", "Human", None),
            ("enter", "field", None, "QueryRoot", "Alien", None),
            ("enter", "name", "alien", "QueryRoot", "Alien", None),
            ("leave", "name", "alien", "QueryRoot", "Alien", None),
            ("enter", "selection_set", None, "Alien", "Alien", None),
            ("enter", "field", None, "Alien", "String!", None),
            ("enter", "name", "__typename", "Alien", "String!", None),
            ("leave", "name", "__typename", "Alien", "String!", None),
            ("leave", "field", None, "Alien", "String!", None),
            ("leave", "selection_set", None, "Alien", "Alien", None),
            ("leave", "field", None, "QueryRoot", "Alien", None),
            ("leave", "selection_set", None, "QueryRoot", "QueryRoot", None),
            ("leave", "operation_definition", None, None, "QueryRoot", None),
            ("leave", "document", None, None, None, None),
        ]

    def supports_traversal_of_input_values():
        visited = []

        schema = build_schema(
            """
            input ComplexInput {
              stringListField: [String]
            }
            """
        )
        complex_input_type = schema.get_type("ComplexInput")
        assert complex_input_type is not None
        type_info = TypeInfo(schema, complex_input_type)

        ast = parse_value('{ stringListField: ["foo"] }')

        class TestVisitor(Visitor):
            @staticmethod
            def enter(node: Node, *_args):
                type_ = type_info.get_input_type()
                visited.append(
                    (
                        "enter",
                        node.kind,
                        node.value if isinstance(node, NameNode) else None,
                        str(type_),
                    )
                )

            @staticmethod
            def leave(node: Node, *_args):
                type_ = type_info.get_input_type()
                visited.append(
                    (
                        "leave",
                        node.kind,
                        node.value if isinstance(node, NameNode) else None,
                        str(type_),
                    )
                )

        visit(ast, TypeInfoVisitor(type_info, TestVisitor()))

        assert visited == [
            ("enter", "object_value", None, "ComplexInput"),
            ("enter", "object_field", None, "[String]"),
            ("enter", "name", "stringListField", "[String]"),
            ("leave", "name", "stringListField", "[String]"),
            ("enter", "list_value", None, "String"),
            ("enter", "string_value", None, "String"),
            ("leave", "string_value", None, "String"),
            ("leave", "list_value", None, "String"),
            ("leave", "object_field", None, "[String]"),
            ("leave", "object_value", None, "ComplexInput"),
        ]

    def supports_traversal_of_selection_sets():
        visited = []

        human_type = test_schema.get_type("Human")
        assert human_type is not None
        type_info = TypeInfo(test_schema, human_type)

        ast = parse("{ name, pets { name } }")
        operation_node = ast.definitions[0]
        assert isinstance(operation_node, OperationDefinitionNode)

        class TestVisitor(Visitor):
            @staticmethod
            def enter(node: Node, *_args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                visited.append(
                    (
                        "enter",
                        node.kind,
                        node.value if isinstance(node, NameNode) else None,
                        str(parent_type),
                        str(type_),
                    )
                )

            @staticmethod
            def leave(node: Node, *_args):
                parent_type = type_info.get_parent_type()
                type_ = type_info.get_type()
                visited.append(
                    (
                        "leave",
                        node.kind,
                        node.value if isinstance(node, NameNode) else None,
                        str(parent_type),
                        str(type_),
                    )
                )

        visit(operation_node.selection_set, TypeInfoVisitor(type_info, TestVisitor()))

        assert visited == [
            ("enter", "selection_set", None, "Human", "Human"),
            ("enter", "field", None, "Human", "String"),
            ("enter", "name", "name", "Human", "String"),
            ("leave", "name", "name", "Human", "String"),
            ("leave", "field", None, "Human", "String"),
            ("enter", "field", None, "Human", "[Pet]"),
            ("enter", "name", "pets", "Human", "[Pet]"),
            ("leave", "name", "pets", "Human", "[Pet]"),
            ("enter", "selection_set", None, "Pet", "[Pet]"),
            ("enter", "field", None, "Pet", "String"),
            ("enter", "name", "name", "Pet", "String"),
            ("leave", "name", "name", "Pet", "String"),
            ("leave", "field", None, "Pet", "String"),
            ("leave", "selection_set", None, "Pet", "[Pet]"),
            ("leave", "field", None, "Human", "[Pet]"),
            ("leave", "selection_set", None, "Human", "Human"),
        ]



from typing import cast, Any, Dict, List, Union

from ...error import GraphQLError
from ...language import (
    DirectiveDefinitionNode,
    DirectiveNode,
    FieldNode,
    InputValueDefinitionNode,
    NonNullTypeNode,
    TypeNode,
    VisitorAction,
    SKIP,
    print_ast,
)
from ...type import GraphQLArgument, is_required_argument, is_type, specified_directives
from . import ASTValidationRule, SDLValidationContext, ValidationContext

__all__ = ["ProvidedRequiredArgumentsRule", "ProvidedRequiredArgumentsOnDirectivesRule"]


class ProvidedRequiredArgumentsOnDirectivesRule(ASTValidationRule):
    """Provided required arguments on directives

    A directive is only valid if all required (non-null without a default value)
    arguments have been provided.

    For internal use only.
    """

    context: Union[ValidationContext, SDLValidationContext]

    def __init__(self, context: Union[ValidationContext, SDLValidationContext]):
        super().__init__(context)
        required_args_map: Dict[
            str, Dict[str, Union[GraphQLArgument, InputValueDefinitionNode]]
        ] = {}

        schema = context.schema
        defined_directives = schema.directives if schema else specified_directives
        for directive in cast(List, defined_directives):
            required_args_map[directive.name] = {
                name: arg
                for name, arg in directive.args.items()
                if is_required_argument(arg)
            }

        ast_definitions = context.document.definitions
        for def_ in ast_definitions:
            if isinstance(def_, DirectiveDefinitionNode):
                required_args_map[def_.name.value] = {
                    arg.name.value: arg
                    for arg in filter(is_required_argument_node, def_.arguments or ())
                }

        self.required_args_map = required_args_map

    def leave_directive(self, directive_node: DirectiveNode, *_args: Any) -> None:
        # Validate on leave to allow for deeper errors to appear first.
        directive_name = directive_node.name.value
        required_args = self.required_args_map.get(directive_name)
        if required_args:

            arg_nodes = directive_node.arguments or ()
            arg_node_set = {arg.name.value for arg in arg_nodes}
            for arg_name in required_args:
                if arg_name not in arg_node_set:
                    arg_type = required_args[arg_name].type
                    arg_type_str = (
                        str(arg_type)
                        if is_type(arg_type)
                        else print_ast(cast(TypeNode, arg_type))
                    )
                    self.report_error(
                        GraphQLError(
                            f"Directive '@{directive_name}' argument '{arg_name}'"
                            f" of type '{arg_type_str}' is required,"
                            " but it was not provided.",
                            directive_node,
                        )
                    )


class ProvidedRequiredArgumentsRule(ProvidedRequiredArgumentsOnDirectivesRule):
    """Provided required arguments

    A field or directive is only valid if all required (non-null without a default
    value) field arguments have been provided.
    """

    context: ValidationContext

    def __init__(self, context: ValidationContext):
        super().__init__(context)

    def leave_field(self, field_node: FieldNode, *_args: Any) -> VisitorAction:
        # Validate on leave to allow for deeper errors to appear first.
        field_def = self.context.get_field_def()
        if not field_def:
            return SKIP
        arg_nodes = field_node.arguments or ()

        arg_node_map = {arg.name.value: arg for arg in arg_nodes}
        for arg_name, arg_def in field_def.args.items():
            arg_node = arg_node_map.get(arg_name)
            if not arg_node and is_required_argument(arg_def):
                self.report_error(
                    GraphQLError(
                        f"Field '{field_node.name.value}' argument '{arg_name}'"
                        f" of type '{arg_def.type}' is required,"
                        " but it was not provided.",
                        field_node,
                    )
                )

        return None


def is_required_argument_node(arg: InputValueDefinitionNode) -> bool:
    return isinstance(arg.type, NonNullTypeNode) and arg.default_value is None
